from astropy.io import fits
import os

import corgidrp.data as data
import corgidrp.mocks as mocks
from corgidrp.l1_to_l2a import detect_cosmic_rays
from corgidrp.detector import find_plateaus, calc_sat_fwc

import numpy as np
from astropy.time import Time
from pathlib import Path
from scipy.ndimage import median_filter
from pytest import approx

###########################################
### Create a dummy non-linearity file ####
#Create a mock dataset because it is a required input when creating a NonLinearityCalibration
dummy_dataset = mocks.create_prescan_files()

# Make a non-linearity correction calibration file
input_non_linearity_filename = "nonlin_table_TVAC.txt"
input_non_linearity_path = os.path.join(os.path.dirname(__file__), "test_data", input_non_linearity_filename)
test_non_linearity_filename = input_non_linearity_filename.split(".")[0] + ".fits"
nonlin_fits_filepath = os.path.join(os.path.dirname(__file__), "test_data", test_non_linearity_filename)
tvac_nonlin_data = np.genfromtxt(input_non_linearity_path, delimiter=",")

pri_hdr, ext_hdr = mocks.create_default_calibration_product_headers()
non_linearity_correction = data.NonLinearityCalibration(tvac_nonlin_data,pri_hdr=pri_hdr,ext_hdr=ext_hdr,input_dataset = dummy_dataset)
non_linearity_correction.save(filename = nonlin_fits_filepath)

# Make a dummy kgain calibration file
kgain = 8.7
gain_value = np.array([[kgain]])
k_gain = data.KGain(gain_value, pri_hdr = pri_hdr, ext_hdr = ext_hdr, input_dataset = dummy_dataset)

# Metadata code from https://github.com/roman-corgi/cgi_iit_drp/blob/main/proc_cgi_frame_NTR/proc_cgi_frame/read_metadata.py

class ReadMetadataException(Exception):
    """Exception class for read_metadata module."""

# Set up to allow the metadata.yaml in the repo be the default
here = Path(os.path.dirname(os.path.abspath(__file__)))
meta_path = Path(here,'test_data','metadata.yaml')

class Metadata(object):
    """ II&T pipeline class to store metadata.

    B Nemati and S Miller - UAH - 03-Aug-2018

    Args:
        meta_path (str): Full path of metadta yaml.

    """

    def __init__(self, meta_path=meta_path):
        self.meta_path = meta_path

        #self.data = self.get_data()
        # self.frame_rows = self.data['frame_rows']
        # self.frame_cols = self.data['frame_cols']
        # self.geom = self.data['geom']

    ### not actually used here, so comment it out to avoid having to require yaml as an import
    # def get_data(self):
    #     """Read yaml data into dictionary.

    #     Returns:
    #         data (dict): Metadata dictionary.
    #     """
    #     with open(self.meta_path, 'r') as stream:
    #         data = yaml.safe_load(stream)
    #     return data

    def slice_section(self, frame, key):
        """Slice 2d section out of frame.

        Args:
            frame (array_like):
                Full frame consistent with size given in frame_rows, frame_cols.
            key (str):
                Keyword referencing section to be sliced; must exist in geom.

        Returns:
            section (array_like): Section of frame
        """
        rows, cols, r0c0 = self._unpack_geom(key)

        section = frame[r0c0[0]:r0c0[0]+rows, r0c0[1]:r0c0[1]+cols]
        if section.size == 0:
            raise ReadMetadataException('Corners invalid')
        return section

    def _unpack_geom(self, key):
        """Safely check format of geom sub-dictionary and return values.

        Args:
            key (str): Keyword referencing section to be sliced; must exist in geom.

        Returns:
            rows (int): Number of rows in section.
            cols (int): Number of columns in section.
            r0c0 (tuple): Initial row and column of section.
        """
        coords = self.geom[key]
        rows = coords['rows']
        cols = coords['cols']
        r0c0 = coords['r0c0']

        return rows, cols, r0c0

    #added in from MetadataWrapper
    def _imaging_area_geom(self):
        """Return geometry of imaging area in reference to full frame.

        Returns:
            rows_im (int): Number of rows corresponding to image frame.
            cols_im (int): Number of columns in section.
            r0c0_im (tuple): Initial row and column of section.
        """

        _, cols_pre, _ = self._unpack_geom('prescan')
        _, cols_serial_ovr, _ = self._unpack_geom('serial_overscan')
        rows_parallel_ovr, _, _ = self._unpack_geom('parallel_overscan')
        #_, _, r0c0_image = self._unpack_geom('image')
        fluxmap_rows, _, r0c0_image = self._unpack_geom('image')

        rows_im = self.frame_rows - rows_parallel_ovr
        cols_im = self.frame_cols - cols_pre - cols_serial_ovr
        r0c0_im = r0c0_image.copy()
        r0c0_im[0] = r0c0_im[0] - (rows_im - fluxmap_rows)

        return rows_im, cols_im, r0c0_im

    def imaging_slice(self, frame):
        """Select only the real counts from full frame and exclude virtual.

        Use this to transform mask and embed from acting on the full frame to
        acting on only the image frame.

        Args:
            frame (array_like):
                Full frame consistent with size given in frame_rows, frame_cols.

        Returns:
            slice (array_like):
                Science image area of full frame.
        """
        rows, cols, r0c0 = self._imaging_area_geom()

        slice = frame[r0c0[0]:r0c0[0]+rows, r0c0[1]:r0c0[1]+cols]

        return slice


## Copy-pasted II&T code from https://github.com/roman-corgi/cgi_iit_drp/blob/main/proc_cgi_frame_NTR/proc_cgi_frame/gsw_remove_cosmics.py ##

def find_plateaus_iit(streak_row, fwc, sat_thresh, plat_thresh, cosm_filter):
    """Find the beginning index of each cosmic plateau in a row.

    Note that i_beg is set at one pixel before first plateau pixel, as these
    pixels immediately neighboring the cosmic plateau are very often affected
    by the cosmic hit as well.

    Args:
        streak_row (float):
            Row with possible cosmics.
        fwc (float):
            Full well capacity of detector *in DNs*.  Note that this may require a
            conversion as FWCs are usually specified in electrons, but the image
            is in DNs at this point.
        sat_thresh (float):
            Multiplication factor for fwc that determines saturated cosmic pixels.
        plat_thresh (float):
            Multiplication factor for fwc that determines edges of cosmic plateu.
        cosm_filter (int):
            Minimum length in pixels of cosmic plateus to be identified.

    Returns:
        array_like, int:
            Index of plateau beginnings, or None if there is no plateau.

    """
    # Lowpass filter row to differentiate plateaus from standalone pixels
    # The way median_filter works, it will find cosmics that are cosm_filter-1
    # wide. Add 1 to cosm_filter to correct for this
    filtered = median_filter(streak_row, cosm_filter+1, mode='nearest')
    saturated = (filtered >= sat_thresh*fwc).nonzero()[0]

    if len(saturated) > 0:
        i_begs = np.array([])
        for i in range(len(saturated)):
            i_beg = saturated[i]
            while i_beg > 0 and streak_row[i_beg] >= plat_thresh*fwc:
                i_beg -= 1
            # unless saturated at col 0, shifts forward 1 to plateau start
            if streak_row[i_beg] < plat_thresh*fwc:
                i_beg += 1
            i_begs = np.append(i_begs, i_beg)

        return np.unique(i_begs).astype(int)
    else:
        return None

def remove_cosmics_iit(image, fwc, sat_thresh, plat_thresh, cosm_filter, cosm_box,
                   cosm_tail, meta=None, mode='image'):
    """Identify and remove saturated cosmic ray hits and tails.

    Use sat_thresh (interval 0 to 1) to set the threshold above which cosmics
    will be detected. For example, sat_thresh=0.99 will detect cosmics above
    0.99*fwc.

    Use plat_thresh (interval 0 to 1) to set the threshold under which cosmic
    plateaus will end. For example, if plat_thresh=0.85, once a cosmic is
    detected the beginning and end of its plateau will be determined where the
    pixel values drop below 0.85*fwc.

    Use cosm_filter to determine the smallest plateaus (in pixels) that will
    be identified. A reasonable value is 2.

    Args:
        image (array_like, float):
            Image area of frame (bias of zero).
        fwc (float):
            Full well capacity of detector *in DNs*.  Note that this may require a
            conversion as FWCs are usually specified in electrons, but the image
            is in DNs at this point.
        sat_thresh (float):
            Multiplication factor for fwc that determines saturated cosmic pixels.
        plat_thresh (float):
            Multiplication factor for fwc that determines edges of cosmic plateau.
        cosm_filter (int):
            Minimum length in pixels of cosmic plateaus to be identified.
        cosm_box (int):
            Number of pixels out from an identified cosmic head (i.e., beginning of
            the plateau) to mask out.
            For example, if cosm_box is 3, a 7x7 box is masked,
            with the cosmic head as the center pixel of the box.
        cosm_tail (int):
            Number of pixels in the row downstream of the end of a cosmic plateau
            to mask.  If cosm_tail is greater than the number of
            columns left to the end of the row from the cosmic
            plateau, the cosmic masking ends at the end of the row. Defaults to 10.
        meta (Metadata class instance):
            Metadata class instance, which is used to determine whether the
            beginning of a plateau is not in the image area, in which case no 
            cosmic ray masking should occur.  Only relevant when mode is 'full'.
            Defaults to None, in which case masking is allowed anywhere on the 
            input frame.
        mode (string):
            If 'image', an image-area input is assumed, and if the input
            tail length is longer than the length to the end of the image-area row,
            the mask is truncated at the end of the row.
            If 'full', a full-frame input is assumed, and if the input tail length
            is longer than the length to the end of the full-frame row, the masking
            continues onto the next row.  Defaults to 'image'.

    Returns:
        mask (array_like, int):
            Mask for pixels that have been set to zero.

    Notes
    -----
    This algorithm uses a row by row method for cosmic removal. It first finds
    streak rows, which are rows that potentially contain cosmics. It then
    filters each of these rows in order to differentiate cosmic hits (plateaus)
    from any outlier saturated pixels. For each cosmic hit it finds the leading
    ledge of the plateau and kills the plateau (specified by cosm_filter) and
    the tail (specified by cosm_tail).

    |<-------- streak row is the whole row ----------------------->|
     ......|<-plateau->|<------------------tail---------->|.........

    B Nemati and S Miller - UAH - 02-Oct-2018
    Kevin Ludwick - UAH - 2024

    """
    mask = np.zeros(image.shape, dtype=int)
    if meta is not None:
        if not isinstance(meta, Metadata):
            raise Exception('meta must be an instance of the Metadata class.')
    if meta is not None and mode=='full':
        im_num_rows = meta.geom['image']['rows']
        im_num_cols = meta.geom['image']['cols']
        im_starting_row = meta.geom['image']['r0c0'][0]
        im_ending_row = im_starting_row + im_num_rows
        im_starting_col = meta.geom['image']['r0c0'][1]
        im_ending_col = im_starting_col + im_num_cols
    else:
        im_starting_row = 0
        im_ending_row = mask.shape[0] - 1 # - 1 to get the index, not size
        im_starting_col = 0
        im_ending_col = mask.shape[1] - 1 # - 1 to get the index, not size

    # Do a cheap prefilter for rows that don't have anything bright
    max_rows = np.max(image, axis=1)
    i_streak_rows = (max_rows >= sat_thresh*fwc).nonzero()[0]

    for i in i_streak_rows:
        row = image[i]
        if i < im_starting_row or i > im_ending_row:
            continue
        # Find if and where saturated plateaus start in streak row
        i_begs = find_plateaus(row, fwc, sat_thresh, plat_thresh, cosm_filter)

        # If plateaus exist, kill the hit and the tail
        cutoffs = np.array([])
        ex_l = np.array([])
        if i_begs is not None:
            for i_beg in i_begs:
                if i_beg < im_starting_col or i_beg > im_ending_col:
                    continue
                # implement cosm_tail
                if i_beg+cosm_filter+cosm_tail+1 > mask.shape[1]:
                    ex_l = np.append(ex_l,
                            i_beg+cosm_filter+cosm_tail+1-mask.shape[1])
                    cutoffs = np.append(cutoffs, i+1)
                streak_end = int(min(i_beg+cosm_filter+cosm_tail+1,
                                mask.shape[1]))
                mask[i, i_beg:streak_end] = 1
                # implement cosm_box
                # can't have cosm_box appear in non-image pixels
                st_row = max(i-cosm_box, im_starting_row)
                end_row = min(i+cosm_box+1, im_ending_row+1)
                st_col = max(i_beg-cosm_box, im_starting_col)
                end_col = min(i_beg+cosm_box+1, im_ending_col+1)
                mask[st_row:end_row, st_col:end_col] = 1
                pass

        if mode == 'full' and len(ex_l) > 0:
            mask_rav = mask.ravel()
            for j in range(len(ex_l)):
                row = cutoffs[j]
                rav_ind = int(row * mask.shape[1] - 1)
                mask_rav[rav_ind:rav_ind + int(ex_l[j])] = 1

    return mask

## Run tests ##

###### create simulated data
datadir = os.path.join(os.path.dirname(__file__), "simdata")

detector_params = data.DetectorParams({}, date_valid=Time("2023-11-01 00:00:00"))

def test_iit_vs_corgidrp():
    """
    Generate mock raw data ('SCI' & 'ENG') and pass into prescan processing function.
    Check output dataset shapes, maintain pointers in the Dataset and Image class,
    and check that output is consistent with results II&T code.
    """

    fwc_em = detector_params.params['FWC_EM_E'] / kgain
    fwc_pp = detector_params.params['FWC_PP_E'] / kgain
    em_gain = 500

    fwc = np.min([fwc_em, fwc_pp*em_gain])
    sat_thresh = 0.99
    plat_thresh = 0.85
    cosm_filter = 2
    cosm_box = 3
    cosm_tail = 10
    mode = 'image'

    # create simulated data
    dataset = mocks.create_cr_dataset(nonlin_fits_filepath, filedir=datadir, numfiles=2,em_gain=em_gain, numCRs=5, plateau_length=10)

    iit_masks = []


    # II&T version
    for frame in dataset:

        cr_mask = remove_cosmics_iit(frame.data, fwc=fwc,
                                 sat_thresh=sat_thresh,
                                 plat_thresh=plat_thresh,
                                 cosm_filter=cosm_filter,
                                 cosm_box=cosm_box,
                                 cosm_tail=cosm_tail,
                                 mode=mode
                                 )

        iit_masks.append(cr_mask)
    iit_masks_arr = np.array(iit_masks)

    # corgidrp version
    crmasked_dataset = detect_cosmic_rays(dataset, detector_params, k_gain, sat_thresh,
                                          plat_thresh, cosm_filter, cosm_box,
                                          cosm_tail, mode)
    corgi_crmask_bool = np.where(crmasked_dataset.all_dq>0,1,0)

    if not corgi_crmask_bool == approx(iit_masks_arr):
        raise Exception(f'Corgidrp and II&T functions do not result in the same CR masks.')

def test_crs_zeros_frame():
    """Verify detect_cosmics does not break for a frame of all zeros
    (should return all zeros)."""

    tol = 1e-13

    # create simulated data
    dataset = mocks.create_cr_dataset(nonlin_fits_filepath, filedir=datadir, numfiles=2,numCRs=5, plateau_length=10)

    # Overwrite data with zeros
    dataset.all_data[:,:,:] = 0.

    output_dataset = detect_cosmic_rays(dataset, detector_params)

    if output_dataset.all_dq != approx(0,abs=tol):
        raise Exception(f'Operating on all zero frames did not return all zero dq mask.')

def test_correct_headers():
    """
    Asserts "FWC_EM_E", "FWC_PP_E", and "SAT_DN" are tracked in the frame headers.
    """
    # create simulated data
    dataset = mocks.create_cr_dataset(nonlin_fits_filepath, filedir=datadir, numfiles=2,numCRs=5, plateau_length=10)
    output_dataset = detect_cosmic_rays(dataset, detector_params, k_gain)

    for frame in output_dataset:
        if not ("FWC_EM_E" in frame.ext_hdr):
            raise Exception("'FWC_EM_E' missing from frame header.")

        if not ("FWC_PP_E" in frame.ext_hdr):
            raise Exception("'FWC_PP_E' missing from frame header.")

        if not ("SAT_DN" in frame.ext_hdr):
            raise Exception("'SAT_DN' missing from frame header.")

def test_saturation_calc():
    """
    Asserts that FWC saturation threshold is calculated correctly.
    """
    sat_thresh = 0.99

    # fwc_em = fwc_pp * em_gain
    fwc_em = np.array([90000,90000])
    fwc_pp = np.array([90,90])
    em_gain = np.array([1000,1000])
    sat_fwcs = calc_sat_fwc(em_gain,fwc_pp,fwc_em,sat_thresh)

    expected = fwc_em * sat_thresh
    if not sat_fwcs == approx(expected):
        raise Exception(f"Saturation full-well capacity calculation incorrect when fwc_em = fwc_pp * em_gain. \nReturned {sat_fwcs} when {expected} was expected.")

    # fwc_em < fwc_pp * em_gain
    fwc_em = np.array([50000,50000])
    fwc_pp = np.array([90,90])
    em_gain = np.array([1000,1000])
    sat_fwcs = calc_sat_fwc(em_gain,fwc_pp,fwc_em,sat_thresh)

    expected = fwc_em * sat_thresh
    if not sat_fwcs == approx(expected):
        raise Exception(f"Saturation full-well capacity calculation incorrect when fwc_em < fwc_pp * em_gain. \nReturned {sat_fwcs} when {expected} was expected.")

    # fwc_em > fwc_pp * em_gain
    fwc_em = np.array([90000,90000])
    fwc_pp = np.array([50,50])
    em_gain = np.array([1000,1000])
    sat_fwcs = calc_sat_fwc(em_gain,fwc_pp,fwc_em,sat_thresh)

    expected = fwc_pp * em_gain * sat_thresh
    if not sat_fwcs == approx(expected):
        raise Exception(f"Saturation full-well capacity calculation incorrect when fwc_em > fwc_pp * em_gain. \nReturned {sat_fwcs} when {expected} was expected.")

    # fwc_em > fwc_pp * em_gain for first frame, fwc_em < fwc_pp * em_gain for second frame
    fwc_em = np.array([90000,50000])
    fwc_pp = np.array([500,90])
    em_gain = np.array([100,1000])
    sat_fwcs = calc_sat_fwc(em_gain,fwc_pp,fwc_em,sat_thresh)

    expected = np.array([49500.,49500.])
    if not sat_fwcs == approx(expected):
        raise Exception(f"Saturation full-well capacity calculation incorrect when frames have different fwc_em, fwc_pp, em_gain. \nReturned {sat_fwcs} when {expected} was expected.")


## Useful constructs from JPL II&T unit tests:

fwc = detector_params.params['FWC_PP_E']
fwcem = detector_params.params['FWC_EM_E']
cosm_filter = 2
plat_thresh = 0.85
sat_thresh = 0.99

# Create a bias subtracted image with cosmics that cover all corner cases
# Make a variety of plateaus
p_basic = np.array([fwc]*cosm_filter)  # Smallest allowed through filter
p_small = np.array([fwc]*(cosm_filter-1))  # Smaller than filter
p_large = np.array([fwc]*cosm_filter*10)  # Larger than filter
p_dip = np.append(p_basic, [plat_thresh*fwc, fwc])  # Shallow dip mid cosmic
p_dip_deep = np.hstack((p_basic, [0.], p_basic))  # Deep dip mid cosmic
p_uneven = np.array([fwc*sat_thresh, fwc, fwc*plat_thresh, fwc,
                     fwc*plat_thresh])  # Uneven cosmic
p_below_min = np.array([fwc*sat_thresh - 1]*cosm_filter)  # Below min value

# Create tail
# An exponential tail with no noise should be able to be perfectly removed
tail = np.exp(np.linspace(0, -10, 50)) * 0.1*fwc

# Create bias subtracted image

bs_dataset = mocks.create_cr_dataset(nonlin_fits_filepath, datadir,numfiles=1,numCRs=0)
bs_dataset.all_data[:,:,:] = 1.
im_width = bs_dataset.all_data.shape[-1]
i_streak_rows_t = np.array([0, im_width//2-1, im_width//2, im_width-1])
cosm_bs = np.append(p_basic, tail)
not_cosm_bs = np.append(p_below_min, tail)

streak_row = np.ones(bs_dataset.all_data.shape[-1])

bs_dataset_below_thresh = bs_dataset.copy()
bs_dataset_single_pix = bs_dataset.copy()
bs_dataset_two_cosm = bs_dataset.copy()

bs_dataset.all_data[0,i_streak_rows_t[0], 0:len(cosm_bs)] = cosm_bs
bs_dataset.all_data[0,i_streak_rows_t[1], 50:50+len(cosm_bs)] = cosm_bs
bs_dataset.all_data[0,i_streak_rows_t[1], 50+len(cosm_bs):50+len(cosm_bs)*2] = cosm_bs
bs_dataset.all_data[0,i_streak_rows_t[2], 51:51+len(cosm_bs)] = cosm_bs
bs_dataset.all_data[0,i_streak_rows_t[3], im_width-len(p_basic):] = p_basic
bs_dataset_below_thresh.all_data[0,im_width//2, 50:50+len(not_cosm_bs)] = not_cosm_bs
bs_dataset_single_pix.all_data[0,im_width//2, im_width//2] = fwc


def test_mask():
    """Assert correct elements are masked."""
    dataset = mocks.create_cr_dataset(nonlin_fits_filepath, filedir=datadir, numfiles=1,numCRs=0, plateau_length=10)
    dataset.all_data[:,:,:] = 1.
    dataset.all_data[0,1, 2:2+len(cosm_bs)] = cosm_bs
    check_mask = np.zeros_like(dataset.all_dq, dtype=int)
    c_tail = 6
    check_mask[0,1, 2:2+cosm_filter+c_tail+1] = 1 #add 1 to include last column in the slice
    dataset_masked = detect_cosmic_rays(dataset, detector_params, k_gain, sat_thresh, plat_thresh, cosm_filter, cosm_box=0, cosm_tail=c_tail)
    if not np.where(dataset_masked.all_dq>0,1,0) == approx(check_mask):
        raise Exception("Incorrect pixels were masked.")

def test_mask_box():
    """Assert correct elements are masked, including the box around
    the cosmic head and the specified cosmic tail."""
    bs_image = np.ones((len(streak_row), 1000))
    bs_image_box = np.ones(bs_image.shape)
    bs_image_box[i_streak_rows_t[1], 50:50+len(cosm_bs)] = cosm_bs
    # these pixels surrounding the cosmic head would not get masked
    # unless cosm_box > 0; doesn't form a full box,
    # but the whole box should get masked
    bs_image_box[i_streak_rows_t[1]-2:i_streak_rows_t[1], 50-2:50+2+1] = 0.1*fwc #not big enough value
    check_mask = np.zeros_like(bs_image, dtype=int)
    check_mask[i_streak_rows_t[1]-2:i_streak_rows_t[1]+2+1,
                50-2:50+2+1] = 1
    # choose cosm_tail >= effective length of simulated tail
    # using cosm_filter=2 and cosm_tail=20:
    check_mask[i_streak_rows_t[1], 50:50+2+20+1] = 1
    check_mask = check_mask.astype(int)
    prihdr, exthdr = mocks.create_default_L1_headers()
    frame = data.Image(bs_image_box, pri_hdr=prihdr,
                    ext_hdr=exthdr)
    dataset = data.Dataset([frame])
    dataset_masked = detect_cosmic_rays(dataset, detector_params, k_gain, sat_thresh,
                        plat_thresh, cosm_filter=2, cosm_box=0,
                        cosm_tail=20)

    assert not (np.array_equal(np.where(dataset_masked.all_dq>0,1,0)[0], check_mask)) # since cosm_box=0

    # now use cosm_box=2 to catch pixels surrounding head
    dataset_masked2 = detect_cosmic_rays(dataset, detector_params, k_gain, sat_thresh,
                        plat_thresh, cosm_filter=2, cosm_box=2,
                        cosm_tail=20)

    assert (np.array_equal(np.where(dataset_masked2.all_dq>0,1,0)[0], check_mask))


def test_mask_box_corners():
    """Assert correct elements are masked, including the box around
    the cosmic head, when cosmic heads appear in corners."""
    check_mask = np.zeros((10,10), dtype=int)
    image = np.zeros((10,10), dtype=float)
    # lower left corner (head #1)
    image[-1,0:4] = fwc
    # near lower left corner (head #2)
    image[-2,1:4] = fwc
    # upper right corner (head #3)
    image[0,-1] = fwc

    # cosmic head #1
    check_mask[-1,0:] = 1
    # tries for a 2x2 box around head in corner
    check_mask[-3:,0:2] = 1
    # cosmic head #2
    check_mask[-2,1:] = 1
    # tries for a 2x2 box around head
    check_mask[-4:,0:4] = 1
    # cosmic head #3 and attempted box around it
    check_mask[0:3,-3:] = 1
    prihdr, exthdr = mocks.create_default_L1_headers()
    frame = data.Image(image, pri_hdr=prihdr,
                    ext_hdr=exthdr)
    dataset = data.Dataset([frame])
    dataset_masked = detect_cosmic_rays(dataset, detector_params, k_gain, sat_thresh,
                        plat_thresh, cosm_filter=2, cosm_box=2)

    if not np.array_equal(np.where(dataset_masked.all_dq>0,1,0)[0], check_mask):
        raise Exception("Incorrect pixels were masked.")

def test_cosm_tail_2():
    """Assert correct elements are masked when 2 cosmic rays are in
    a single row.  cosm_box=0 for simplicity."""
    check_mask = np.zeros((10,10), dtype=int)
    image = np.zeros((10,10), dtype=float)
    # head #1
    image[-2,0:4] = fwc
    # head #2
    image[-2,6:9] = fwc

    # for cosm_filter=2 and cosm_tail=1:
    # head #1
    check_mask[-2,0:0+2+1+1] = 1
    # cosmic head #2
    check_mask[-2,6:6+2+1+1] = 1
    prihdr, exthdr = mocks.create_default_L1_headers()
    frame = data.Image(image, pri_hdr=prihdr,
                    ext_hdr=exthdr)
    dataset = data.Dataset([frame])
    dataset_masked = detect_cosmic_rays(dataset, detector_params, k_gain, sat_thresh,
                        plat_thresh, cosm_filter=2, cosm_box=0,
                        cosm_tail=1)

    assert (np.array_equal(np.where(dataset_masked.all_dq>0,1,0)[0], check_mask))

    # for cosm_filter=2 and cosm_tail=3 (overlap due to masked tails):
    # head #1
    check_mask[-2,0:0+2+3+1] = 1
    # cosmic head #2
    check_mask[-2,6:6+2+3+1] = 1

    dataset_masked = detect_cosmic_rays(dataset, detector_params, k_gain, sat_thresh,
                        plat_thresh, cosm_filter=2, cosm_box=0,
                        cosm_tail=3)

    assert (np.array_equal(np.where(dataset_masked.all_dq>0,1,0)[0], check_mask))

def test_cosm_tail_bleed_over():
    """Assert correct elements are masked when cosmic ray in
        a single row with bleed over into next row, while preventng 
        detections outside of image area from being flagged."""
    im_num_rows = 1024
    im_num_cols = 1024
    im_starting_row = 13
    im_ending_row = im_starting_row + im_num_rows
    im_starting_col = 1088
    im_ending_col = im_starting_col + im_num_cols
    check_mask = np.zeros((1200,2200), dtype=int)
    image = np.zeros((1200,2200), dtype=float)
    # head (not saturating for easy distinguishing b/w cosmic and saturation flag values)
    image[im_ending_row-1,im_ending_col-4:im_ending_col-1] = sat_thresh*fwcem/8.7
    # would normally trigger a detection, but not inside image area:
    image[-2,6:9] = fwc

    # cosmic head
    check_mask[im_ending_row-1,im_ending_col-4:] = 1
    # with cosm_tail=100, and (88-2) left in row after cosm_filter, 
    # so bleed 12-2 over next row
    check_mask[im_ending_row,0:10] = 1
    # cosm_box gets cut short one row since the end of the image area is 
    # reached with only 1 extra row of masking below the cosmic head
    check_mask[im_ending_row-3:im_ending_row+1,
                im_ending_col-6:im_ending_col-1] = 1 # cosm_box=2
    
    prihdr, exthdr = mocks.create_default_L1_headers()
    frame = data.Image(image, pri_hdr=prihdr,
                    ext_hdr=exthdr)
    dataset = data.Dataset([frame])
    dataset_masked = detect_cosmic_rays(dataset, detector_params, k_gain, sat_thresh,
                        plat_thresh, cosm_filter=2, cosm_box=2,
                        cosm_tail=100, mode='full')

    # cosmic ray is found by first finding saturation, so mask due to cosmic rays 
    # could be valued at 128 or 128+32
    assert (np.array_equal(np.where(dataset_masked.all_dq>=128,1,0)[0], check_mask))
    # saturated row 1200-2 outside of image area covered in saturation mask:
    assert (1198 in np.where(dataset_masked.all_dq[0]==32)[0])

def test_i_begs():
    """Verify that function returns correct i_begs result."""
    beg = 50
    streak_row_copy = streak_row.copy()
    streak_row_copy[beg:beg+len(cosm_bs)] = cosm_bs
    i_beg = find_plateaus(streak_row_copy, fwc, sat_thresh,
                            plat_thresh, cosm_filter)
    if not i_beg == beg:
        raise Exception("find_plateaus returned incorrect streak beginning.")

def test_left_edge_i_begs():
    """Verify that function returns correct i_begs result at left edge."""
    beg = 0
    streak_row_copy = streak_row.copy()
    streak_row_copy[beg:len(cosm_bs)] = cosm_bs
    i_beg = find_plateaus(streak_row_copy, fwc, sat_thresh,
                            plat_thresh, cosm_filter)
    if not i_beg == np.array([beg]):
        raise Exception("find_plateaus returned incorrect streak beginning at left edge.")

def test_right_edge_i_begs():
    """Verify that function returns correct i_begs result at right edge."""
    cosm = p_basic
    beg = len(streak_row)-len(cosm)
    streak_row_copy = streak_row.copy()
    streak_row_copy[beg:] = cosm
    i_beg = find_plateaus(streak_row_copy, fwc, sat_thresh,
                            plat_thresh, cosm_filter)
    if not i_beg == np.array([beg]):
        # print(streak_row)
        # print(i_beg,beg)
        raise Exception("find_plateaus returned incorrect streak beginning at right edge.")

def test_two_cosm_i_begs():
    """Verify that function returns correct i_begs result for two cosm."""
    beg1 = len(streak_row) - len(cosm_bs)*2
    beg2 = beg1 + len(cosm_bs)
    streak_row_copy = streak_row.copy()
    streak_row_copy[beg1:beg1 + len(cosm_bs)] = cosm_bs
    streak_row_copy[beg2:beg2 + len(cosm_bs)] = cosm_bs

    i_begs = find_plateaus(streak_row_copy, fwc, sat_thresh,
                            plat_thresh, cosm_filter)
    if not np.array_equal(i_begs, np.array([beg1, beg2])):
        raise Exception("find_plateaus returned incorrect streak beginning for two CRs in one row.")

def test_p_small():
    """Verify that function ignores plateaus smaller than filter size."""
    cosm = np.append(p_small, tail)
    beg = 50
    streak_row_copy = streak_row.copy()
    streak_row_copy[beg:beg+len(cosm)] = cosm
    i_beg = find_plateaus(streak_row_copy, fwc,
                            sat_thresh, plat_thresh,
                            cosm_filter)
    if not i_beg is None:
        raise Exception("find_plateaus did not ignore plateau smaller than filter size.")

def test_p_large():
    """Verify that function returns correct results for large plateaus."""
    cosm = np.append(p_large, tail)
    beg = 50
    streak_row_copy = streak_row.copy()
    streak_row_copy[beg:beg+len(cosm)] = cosm
    i_beg = find_plateaus(streak_row_copy, fwc,
                            sat_thresh, plat_thresh,
                            cosm_filter)
    if not i_beg == np.array([beg]):
        raise Exception("find_plateaus returned incorrect streak beginning for large plateau.")

def test_p_dip():
    """Verify that function still recognizes a plateau with a dip."""
    cosm = np.append(p_dip, tail)
    beg = 50
    streak_row_copy = streak_row.copy()
    streak_row_copy[beg:beg+len(cosm)] = cosm
    i_beg = find_plateaus(streak_row_copy, fwc,
                            sat_thresh, plat_thresh,
                            cosm_filter)
    if not i_beg == np.array([beg]):
        raise Exception("find_plateaus did not recognize plateau with a dip.")

def test_p_dip_deep():
    """Verify that the function recognizes plateau with a single pixel dip
    below plat_thresh and does not set the end at the dip."""
    cosm = np.append(p_dip_deep, tail)
    beg = 50
    streak_row_copy = streak_row.copy()
    streak_row_copy[beg:beg+len(cosm)] = cosm
    i_beg = find_plateaus(streak_row_copy, fwc,
                            sat_thresh, plat_thresh,
                            cosm_filter)
    if not i_beg[0] == beg:
        raise Exception("find_plateaus returned incorrect result for plateau with a deep dip.")
    # also finds where the dip is when cosm_filter=2, and the dip is
    # 2 away, which is 1 before the next plateau
    if not i_beg[1] == beg+3:
        raise Exception("find_plateaus returned incorrect result for plateau with a deep dip.")

def test_p_uneven():
    """Verify that function still recognizes an uneven plateau."""
    cosm = np.append(p_uneven, tail)
    beg = 50
    streak_row_copy = streak_row.copy()
    streak_row_copy[beg:beg+len(cosm)] = cosm
    i_beg = find_plateaus(streak_row_copy, fwc,
                            sat_thresh, plat_thresh,
                            cosm_filter)
    if not i_beg == np.array([beg]):
        raise Exception("find_plateaus returned incorrect streak beginning for uneven plateau.")

def test_p_below_min():
    """Verify that function ignores plateaus below saturation thresh."""
    cosm = np.append(p_below_min, tail)
    beg = 50
    streak_row_copy = streak_row.copy()
    streak_row_copy[beg:beg+len(cosm)] = cosm
    i_beg = find_plateaus(streak_row_copy, fwc,
                            sat_thresh, plat_thresh,
                            cosm_filter)
    if not i_beg is None:
        raise Exception("find_plateaus did not ignore plateau below sat threshold.")


if __name__ == "__main__":
    test_iit_vs_corgidrp()
    test_crs_zeros_frame()
    test_correct_headers()
    test_saturation_calc()
    test_mask()
    test_i_begs()
    test_left_edge_i_begs()
    test_right_edge_i_begs()
    test_two_cosm_i_begs()
    test_p_small()
    test_p_large()
    test_p_dip()
    test_p_dip_deep()
    test_p_uneven()
    test_p_below_min()
    test_mask_box()
    test_mask_box_corners()
    test_cosm_tail_2()
    test_cosm_tail_bleed_over()


import os
import glob
import numpy as np
import corgidrp.data as data
import corgidrp.mocks as mocks
import corgidrp.detector as detector
import corgidrp.flat as flat
from corgidrp.bad_pixel_calibration import create_bad_pixel_map
from corgidrp.darks import build_trad_dark
from corgidrp import default_cal_dir
import re

np.random.seed(456)

# Get the flag to bit map and flag to value map
FLAG_TO_BIT_MAP = data.get_flag_to_bit_map()
FLAG_TO_VALUE_MAP = data.get_flag_to_value_map()

def generate_badpixel_map(datadir=None, dthresh=6):
    """
    Create simulated dark and flat calibration data, inject hot and dead pixels,
    and then generate a bad pixel map. Returns the bad pixel map,
    the dark frame, and the flat field frame.
    
    Args:
        datadir (str, optional): Directory to store simulated data. If None,
                                 a "simdata" folder in the current directory is used.
        dthresh (int, optional): Threshold parameter to pass into create_bad_pixel_map.
                                 Defaults to 6.
    
    Returns:
        tuple: (badpixelmap, dark_frame, flat_frame)
    """
    # Set the data directory.
    if datadir is None:
        datadir = os.path.join(os.path.dirname(__file__), "simdata")
    os.makedirs(datadir, exist_ok=True)
    
    # --- Create and load dark data ---
    mocks.create_dark_calib_files(filedir=datadir)
    dark_filenames = glob.glob(os.path.join(datadir, "simcal_dark*.fits"))
    dark_dataset = data.Dataset(dark_filenames)
    
    # Create the dark frame using default detector parameters.
    detector_params = data.DetectorParams({})
    dark_frame = build_trad_dark(dark_dataset, detector_params)
    
    # Inject "hot" pixels into the dark frame.
    col_hot_pixels = [12, 123, 234, 456, 678, 890]
    row_hot_pixels = [546, 789, 123, 43, 547, 675]
    for i_col in col_hot_pixels:
        for i_row in row_hot_pixels:
            dark_frame.data[i_col, i_row] = 300

    # --- Create and load flat data ---
    mocks.create_simflat_dataset(filedir=datadir)
    simflat_filenames = glob.glob(os.path.join(datadir, "sim_flat*.fits"))
    simflat_dataset = data.Dataset(simflat_filenames)
    
    # Create a dummy flat field (flat division) image.
    mocks.create_flatfield_dummy(filedir=datadir)
    flat_filenames = glob.glob(os.path.join(datadir, "flat_field*.fits"))
    flat_dataset = data.Dataset(flat_filenames)
    flat_frame = flat.create_flatfield(flat_dataset)
    
    # Inject "dead" pixels into the flat field image.
    col_dead_pixels = [12, 120, 234, 450, 678, 990]
    row_dead_pixels = [546, 89, 123, 243, 447, 675]
    for i_col in col_dead_pixels:
        for i_row in row_dead_pixels:
            flat_frame.data[i_col, i_row] = 0.3

    # --- Create the bad pixel map ---
    bpm = create_bad_pixel_map(flat_dataset, dark_frame, flat_frame, dthresh=dthresh)
    
    return bpm, dark_frame, flat_frame


def test_badpixelmap(): 
    '''

    Tests the creation of badpixelmaps.

    Create master darks and master flats, inject some hot and cold pixels
    to create a master bad pixel map. 

    '''

    ###### make the badpixel map (input the flat_dataset just as a dummy):
    badpixelmap, dark_frame, flat_frame = generate_badpixel_map()

    # # Use np.unpackbits to unpack the bits - big endien demical to binary
    badpixelmap_bits = data.unpackbits_64uint(badpixelmap.data[:, :, np.newaxis], axis=2)  # unit64 to binary
    badpixelmap_repacked = data.packbits_64uint(badpixelmap_bits, axis=2).reshape(badpixelmap.data.shape)
    assert np.array_equal(badpixelmap_repacked, badpixelmap.data) # check if the repacked data is the same as the original

    # Checking that everywhere there's a badpixel is in one of the two lists
    bp_locations = np.argwhere(badpixelmap.data)
    col_hot = [12, 123, 234, 456, 678, 890]
    col_dead = [12, 120, 234, 450, 678, 990]
    row_hot = [546, 789, 123, 43, 547, 675]
    row_dead = [546, 89, 123, 243, 447, 675]
    
    for (i_col, i_row) in bp_locations:
        assert (i_col in col_hot or i_col in col_dead), f"Column {i_col} not expected."
        assert (i_row in row_hot or i_row in row_dead), f"Row {i_row} not expected."

    # Checking that hot pixels are at the expected locations - bit #3
    hot_pixel_bit_position = 63 - FLAG_TO_BIT_MAP["hot_pixel"]
    hot_pixel_locations = np.where(badpixelmap_bits[:,:,hot_pixel_bit_position])
    for ii in hot_pixel_locations[0]:
        assert ii in col_hot, f"Hot pixel at column {ii} not expected."
    for jj in hot_pixel_locations[1]:
        assert jj in row_hot, f"Hot pixel at row {jj} not expected."

    # Checking that CR are at the expected locations - bit #2
    dead_pixel_bit_position = 63 - FLAG_TO_BIT_MAP["bad_pixel"]
    dead_pixel_locations = np.where(badpixelmap_bits[:,:,dead_pixel_bit_position])
    for ii in dead_pixel_locations[0]:
        assert ii in col_dead, f"Dead pixel at column {ii} not expected." 
    for jj in dead_pixel_locations[1]:
       assert jj in row_dead, f"Dead pixel at row {jj} not expected."


def test_packing_unpacking_uint64():
    '''
    Test the packing and unpacking of uint64 data
    ''' 

    # Checking whether we can assign demical higher than 255 to the badpixelmap to test our function for bit unpacking and packing
    packed = np.zeros((3, 3), dtype='>u8')

    # Get the bit position for 'TBD'
    bit_position = FLAG_TO_BIT_MAP["TBD"]
    flag_value = FLAG_TO_VALUE_MAP["TBD"]

    packed[0, 0] = flag_value

    unpacked_bits = data.unpackbits_64uint(packed[:, :, np.newaxis], axis=2)

    # Compute the expected unpacked index for 'big' endian
    expected_index = 63 - bit_position

    # Check the expected bit index is set
    assert unpacked_bits[0, 0, expected_index] == 1, f"Bit {bit_position} should be at index {expected_index}"

    # Check that only one bit is set
    assert np.sum(unpacked_bits[0, 0]) == 1, "Only one bit should be set"


def test_output_filename_convention():
    print("**Testing output filename naming conventions**")

    bp_fake_data = np.array([
        [0,  1,  0],   
        [1,  0, 0],  
        [0, 0, 0]
    ], dtype=float)

    # Create a fake input dataset to set the filename
    input_prihdr, input_exthdr = mocks.create_default_L2b_headers()
    fake_input_image = data.Image(bp_fake_data, pri_hdr=input_prihdr, ext_hdr=input_exthdr)
    fake_input_image.filename = f"CGI_{input_prihdr['VISITID']}_{data.format_ftimeutc(input_exthdr['FTIMEUTC'])}_L2b.fits"
    fake_input_dataset = data.Dataset(frames_or_filepaths=[fake_input_image, fake_input_image])

    bpcal_prihdr, bpcal_exthdr = mocks.create_default_calibration_product_headers()
    
    badpixelmap = data.BadPixelMap(bp_fake_data, pri_hdr=bpcal_prihdr, ext_hdr=bpcal_exthdr, input_dataset=fake_input_dataset)
    badpixelmap.save(filedir=default_cal_dir)

     # Construct the expected filename from the last input dataset filename.
    expected_filename = re.sub('_L[0-9].', '_BPM_CAL', fake_input_dataset[-1].filename)
    full_expected_path = os.path.join(default_cal_dir, expected_filename)
    
    assert os.path.exists(full_expected_path), (
        f"Expected file {expected_filename} not found in {default_cal_dir}."
    )
    print("The bad pixel map calibration product file exists and meets the expected naming convention.")


if __name__ == "__main__":
    test_badpixelmap()
    test_packing_unpacking_uint64()
    test_output_filename_convention()


import glob
import os

import corgidrp
import corgidrp.data as data
from corgidrp.l1_to_l2a import prescan_biassub
import corgidrp.mocks as mocks
from corgidrp.detector import detector_areas, unpack_geom, imaging_area_geom

import numpy as np
import yaml
from astropy.io import fits
from pathlib import Path

from pytest import approx

old_err_tracking = corgidrp.track_individual_errors

# make a mock DetectorNoiseMaps instance (to get the bias offset input)
im_rows, im_cols, _ = unpack_geom('SCI', 'image', detector_areas)
rows = detector_areas['SCI']['frame_rows']
cols = detector_areas['SCI']['frame_cols']

Fd = np.ones((rows, cols))
Dd = 3/3600*np.ones((rows, cols))
Cd = 0.02*np.ones((rows, cols))

Ferr = np.zeros((rows, cols))
Derr = np.zeros((rows, cols))
Cerr = np.zeros((rows, cols))
Fdq = Ferr.copy().astype(int)
Ddq = Derr.copy().astype(int)
Cdq = Cerr.copy().astype(int)
noise_maps = mocks.create_noise_maps(Fd, Ferr, Fdq, Cd,
                                            Cerr, Cdq, Dd, Derr, Ddq)

# Expected output image shapes
shapes = {
    'SCI' : {
        True : (1200,2200),
        False : (1024,1024)
    },
    'ENG' : {
        True: (2200,2200),
        False : (1024,1024)
    }
}

# Copy-pasted II&T code

# Metadata code from https://github.com/roman-corgi/cgi_iit_drp/blob/main/proc_cgi_frame_NTR/proc_cgi_frame/read_metadata.py

class ReadMetadataException(Exception):
    """Exception class for read_metadata module."""

# Set up to allow the metadata.yaml in the repo be the default
here = Path(os.path.dirname(os.path.abspath(__file__)))
meta_path = Path(here,'test_data','metadata.yaml')

class Metadata(object):
    """ II&T pipeline class to store metadata.

    B Nemati and S Miller - UAH - 03-Aug-2018

    Args:
        meta_path (str): Full path of metadta yaml.

    Attributes:
        data (dict):
            Data from metadata file.
        geom (SimpleNamespace):
            Geometry specific data.
    """

    def __init__(self, meta_path=meta_path):
        self.meta_path = meta_path

        self.data = self.get_data()
        self.frame_rows = self.data['frame_rows']
        self.frame_cols = self.data['frame_cols']
        self.geom = self.data['geom']

    def get_data(self):
        """Read yaml data into dictionary.

        Returns:
            data (dict): Metadata dictionary.
        """
        with open(self.meta_path, 'r') as stream:
            data = yaml.safe_load(stream)
        return data

    def slice_section(self, frame, key):
        """Slice 2d section out of frame.

        Args:
            frame (array_like):
                Full frame consistent with size given in frame_rows, frame_cols.
            key (str):
                Keyword referencing section to be sliced; must exist in geom.

        Returns:
            section (array_like): Section of frame
        """
        rows, cols, r0c0 = self._unpack_geom(key)

        section = frame[r0c0[0]:r0c0[0]+rows, r0c0[1]:r0c0[1]+cols]
        if section.size == 0:
            raise ReadMetadataException('Corners invalid')
        return section

    def _unpack_geom(self, key):
        """Safely check format of geom sub-dictionary and return values.

        Args:
            key (str): Keyword referencing section to be sliced; must exist in geom.

        Returns:
            rows (int): Number of rows in section.
            cols (int): Number of columns in section.
            r0c0 (tuple): Initial row and column of section.
        """
        coords = self.geom[key]
        rows = coords['rows']
        cols = coords['cols']
        r0c0 = coords['r0c0']

        return rows, cols, r0c0

    #added in from MetadataWrapper
    def _imaging_area_geom(self):
        """Return geometry of imaging area in reference to full frame.

        Returns:
            rows_im (int): Number of rows corresponding to image frame.
            cols_im (int): Number of columns in section.
            r0c0_im (tuple): Initial row and column of section.
        """

        _, cols_pre, _ = self._unpack_geom('prescan')
        _, cols_serial_ovr, _ = self._unpack_geom('serial_overscan')
        rows_parallel_ovr, _, _ = self._unpack_geom('parallel_overscan')
        #_, _, r0c0_image = self._unpack_geom('image')
        fluxmap_rows, _, r0c0_image = self._unpack_geom('image')

        rows_im = self.frame_rows - rows_parallel_ovr
        cols_im = self.frame_cols - cols_pre - cols_serial_ovr
        r0c0_im = r0c0_image.copy()
        r0c0_im[0] = r0c0_im[0] - (rows_im - fluxmap_rows)

        return rows_im, cols_im, r0c0_im

    def imaging_slice(self, frame):
        """Select only the real counts from full frame and exclude virtual.

        Use this to transform mask and embed from acting on the full frame to
        acting on only the image frame.

        Args:
            frame (array_like):
                Full frame consistent with size given in frame_rows, frame_cols.

        Returns:
            slice (array_like):
                Science image area of full frame.
        """
        rows, cols, r0c0 = self._imaging_area_geom()

        slice = frame[r0c0[0]:r0c0[0]+rows, r0c0[1]:r0c0[1]+cols]

        return slice

# EMCCDFrame code from https://github.com/roman-corgi/cgi_iit_drp/blob/main/proc_cgi_frame_NTR/proc_cgi_frame/gsw_emccd_frame.py#L9

class EMCCDFrameException(Exception):
    """Exception class for emccd_frame module."""

class EMCCDFrame:
    """Get data from EMCCD frame and subtract the bias and bias offset.

    S Miller - UAH - 16-April-2019

    Args:
        frame_dn (array_like):
            Raw EMCCD full frame (DN).
        meta (instance):
            Instance of Metadata class containing detector metadata.
        fwc_em (float):
            Detector EM gain register full well capacity (DN).
        fwc_pp (float):
            Detector image area per-pixel full well capacity (DN).
        em_gain (float):
            Gain from EM gain register, >= 1 (unitless).
        bias_offset (float):
            Median number of counts in the bias region due to fixed non-bias noise
            not in common with the image region.  Basically we compute the bias
            for the image region based on the prescan from each frame, and the
            bias_offset is how many additional counts the prescan had from extra
            noise not captured in the master dark fit.  This value is subtracted
            from each measured bias.  Units of DN.

    Attributes:
        image (array_like):
            Image section of frame (DN).
        prescan (array_like):
            Prescan section of frame (DN).
        al_prescan (array_like):
            Prescan with row numbers relative to the first image row (DN).
        frame_bias (array_like):
            Column vector with each entry the median of the prescan row minus the
            bias offset (DN).
        bias (array_like):
            Column vector with each entry the median of the prescan row relative
            to the first image row minus the bias offset (DN).
        frame_bias0 (array_like):
            Total frame minus the bias (row by row) minus the bias offset (DN).
        image_bias0 (array_like):
            Image area minus the bias (row by row) minus the bias offset (DN).
    """

    def __init__(self, frame_dn, meta, fwc_em, fwc_pp, em_gain, bias_offset):
        self.frame_dn = frame_dn
        self.meta = meta
        self.fwc_em = fwc_em
        self.fwc_pp = fwc_pp
        self.em_gain = em_gain
        self.bias_offset = bias_offset

        # Divide frame into sections
        try:
            self.image = self.meta.slice_section(self.frame_dn, 'image')
            self.prescan = self.meta.slice_section(self.frame_dn, 'prescan')
        except Exception:
            raise EMCCDFrameException('Frame size inconsistent with metadata')

        # Get the part of the prescan that lines up with the image, and do a
        # row-by-row bias subtraction on it
        i_r0 = self.meta.geom['image']['r0c0'][0]
        p_r0 = self.meta.geom['prescan']['r0c0'][0]
        i_nrow = self.meta.geom['image']['rows']
        # select the good cols for getting row-by-row bias
        st = self.meta.geom['prescan']['col_start']
        end = self.meta.geom['prescan']['col_end']
        # over all prescan rows
        medbyrow_tot = np.median(self.prescan[:,st:end], axis=1)[:, np.newaxis]
        # prescan relative to image rows
        self.al_prescan = self.prescan[(i_r0-p_r0):(i_r0-p_r0+i_nrow), :]
        medbyrow = np.median(self.al_prescan[:,st:end], axis=1)[:, np.newaxis]

        # Get data from prescan (image area)
        self.bias = medbyrow - self.bias_offset
        self.image_bias0 = self.image - self.bias

        # over total frame
        self.frame_bias = medbyrow_tot - self.bias_offset
        self.frame_bias0 = self.frame_dn[p_r0:, :] -  self.frame_bias

# Run tests

def test_prescan_sub():
    """
    Generate mock raw data ('SCI' & 'ENG') and pass into prescan processing function.
    Check output dataset shapes, maintain pointers in the Dataset and Image class,
    and check that output is consistent with results II&T code.
    """

    tol = 0.01

    ###### create simulated data
    datadir = os.path.join(os.path.dirname(__file__), "simdata")

    for arrtype in ['SCI','ENG']:
        # create simulated data
        dataset = mocks.create_prescan_files(filedir=datadir, arrtype=arrtype)

        filenames = glob.glob(os.path.join(datadir, f"sim_prescan_{arrtype}*.fits"))

        dataset = data.Dataset(filenames)
        assert len(dataset) == 2

        iit_images = []
        iit_frames = []

        # II&T version
        for fname in filenames:

            l1_data = fits.getdata(fname)

            # Read in data
            meta_path = Path(here,'test_data','metadata.yaml') if arrtype == 'SCI' else Path(here,'test_data','metadata_eng.yaml')
            meta = Metadata(meta_path = meta_path)
            frameobj = EMCCDFrame(l1_data,
                                    meta,
                                    1., # fwc_em_dn
                                    1., # fwc_pp_dn
                                    1., # em_gain
                                    0.) # bias_offset

            # Subtract bias and bias offset and get cosmic mask
            iit_images.append(frameobj.image_bias0) # Science area
            iit_frames.append(frameobj.frame_bias0) # Full frame

        if len(dataset) != 2:
            raise Exception(f"Mock dataset is an unexpected length ({len(dataset)}).")

        for return_full_frame in [True, False]:
            for return_imaging_area in [True, False]:
                if return_full_frame is True and return_imaging_area is True:
                    continue # don't test this case b/c function not can't work
                output_dataset = prescan_biassub(dataset, noise_maps, return_full_frame=return_full_frame, use_imaging_area=return_imaging_area)

                # Check that output shape is as expected
                output_shape = output_dataset[0].data.shape
                if return_imaging_area is False:
                    shape_compare = shapes[arrtype][return_full_frame]
                else:
                    r, c, _ = imaging_area_geom(arrtype) 
                    shape_compare = (r,c)
                if output_shape != shape_compare:
                    raise Exception(f"Shape of output frame for {arrtype}, return_full_frame={return_full_frame} is {output_shape}, \nwhen {shapes[arrtype][return_full_frame]} was expected.")

                # Check that bias extension has the right size, dtype
                for i, frame in enumerate(output_dataset):

                    try: 
                        frame_bias = frame.hdu_list['BIAS'].data
                    except KeyError:
                        raise Exception(f"BIAS extension not found in frame {i}.")
                    
                    if frame_bias.shape != (frame.data.shape[0],):
                        raise Exception(f"Bias of frame {i} has shape {frame.bias.shape} when we expected {(frame.data.shape[0],)}.")
                    
                    if frame_bias.dtype != np.float32:
                        raise Exception(f"Bias of frame {i} does not have datatype np.float32.")

                # Check that corgiDRP and II&T pipeline produce the same result
                corgidrp_result = output_dataset[0].data
                if return_imaging_area is True:
                    continue
                iit_result = iit_frames[0] if return_full_frame else iit_images[0]
                if np.nanmax(np.abs(corgidrp_result-iit_result)) > tol:
                    raise Exception(f"corgidrp result does not match II&T result for generated mock data, arrtype={arrtype}, return_full_frame={return_full_frame}.")

                # check that data, err, and dq arrays are consistently modified
                output_dataset.all_data[0, 0, 0] = 0.
                if output_dataset[0].data[0, 0] != 0. :
                    raise Exception("Modifying dataset.all_data did not modify individual frame data.")

                output_dataset[0].data[0,0] = 1.
                if output_dataset.all_data[0,0,0] != 1. :
                    raise Exception("Modifying individual frame data did not modify dataset.all_data.")

                output_dataset.all_err[0, 0, 0, 0] = 0.
                if output_dataset[0].err[0, 0, 0] != 0. :
                    raise Exception("Modifying dataset.all_err did not modify individual frame err.")

                output_dataset[0].err[0, 0, 0] = 1.
                if output_dataset.all_err[0, 0, 0, 0] != 1. :
                    raise Exception("Modifying individual frame err did not modify dataset.all_err.")

                output_dataset.all_dq[0, 0, 0] = 0.
                if output_dataset[0].dq[0, 0] != 0. :
                    raise Exception("Modifying dataset.all_dq did not modify individual frame dq.")

                output_dataset[0].dq[0,0] = 1.
                if output_dataset.all_dq[0,0,0] != 1. :
                    raise Exception("Modifying individual frame dq did not modify dataset.all_dq.")

def test_bias_zeros_frame():
    """Verify prescan_biassub does not break for a frame of all zeros
    (should return all zeros)."""

    tol = 1e-13

    ###### create simulated data
    datadir = os.path.join(os.path.dirname(__file__), "simdata")

    for arrtype in ['SCI', 'ENG']:
        # create simulated data
        dataset = mocks.create_prescan_files(filedir=datadir, arrtype=arrtype,numfiles=1)

        # Overwrite data with zeros
        dataset.all_data[:,:,:] = 0.

        for return_full_frame in [True, False]:

            output_dataset = prescan_biassub(dataset, noise_maps=None, return_full_frame=return_full_frame)

            if np.max(np.abs(output_dataset.all_data)) > tol:
                raise Exception(f'Operating on all zero frame did not return all zero frame.')

            if np.max(np.abs(output_dataset.all_err)) > tol:
                raise Exception(f'Operating on all zero frame did not return all zero error.')           
            
            for i,frame in enumerate(output_dataset):
                try: 
                    frame_bias = frame.hdu_list['BIAS'].data
                except KeyError:
                    raise Exception(f"BIAS extension not found in frame {i}.".format(i))
                
                if np.max(np.abs(frame_bias)) > tol:
                    raise Exception(f'Operating on all zero frame did not return all zero bias.')

def test_bias_hvoff():
    """
    Verify that the function finds bias for gaussian distribution, with no
    contribution from the effect of gain ("hv" is the voltage applied in
    the EM gain register).
    The error tolerance is set by the standard error on the median of
    the Gaussian noise, not the mean.
    """
    corgidrp.track_individual_errors = True # needs to run with error tracking on

    # Set tolerance
    tol = 1.
    err_tol = 0.02
    bval = 100.
    sig = 1.
    seed = 12346

    ###### create simulated data
    datadir = os.path.join(os.path.dirname(__file__), "simdata")

    for arrtype in ['SCI', 'ENG']:
        # create simulated data
        dataset = mocks.create_prescan_files(filedir=datadir, arrtype=arrtype,
                                             numfiles=1)

        # Overwrite data with normal distribution
        rng = np.random.default_rng(seed)
        dataset.all_data[:,:,:] = rng.normal(bval, sig,
                                             size=dataset.all_data.shape)

        for return_full_frame in [True, False]:

            output_dataset = prescan_biassub(dataset, noise_maps, return_full_frame=return_full_frame)

            # Compare bias measurement to expectation
            if np.any(np.abs(output_dataset[0].hdu_list['BIAS'].data - bval) > tol):
                raise Exception(f'Higher than expected error in bias measurement for hvoff distribution.')

            # Compare error to expected standard error of the median
            std_err = sig / np.sqrt(detector_areas[arrtype]['prescan_reliable']['cols']) * np.sqrt(np.pi / 2.)
            if np.max(np.abs(output_dataset[0].err[1]) - std_err) > err_tol:
                raise Exception(f'Higher than expected std. error in bias measurement for hvoff distribution: \n{np.max(np.abs(output_dataset[0].err[1]))} when we expect {std_err} +- {err_tol} ')

    corgidrp.track_individual_errors = old_err_tracking

def test_bias_hvon():
    """
    Verify that the function finds bias for a gaussian distribution, plus
    additional contributions from the effect of gain ("hv" is the voltage
    applied in the EM gain register), approximated as an exponential distribution
    + inflated values for the "unreliable" prescan region, minus the mean of the
    exponential distribution to keep the DC contribution 0.
    Also tests that only the good columns are used for the bias.
    """

    # Set tolerance
    tol = 6.
    bval = 100.
    expmean = 10.
    seed = 12346

    ###### create simulated data
    datadir = os.path.join(os.path.dirname(__file__), "simdata")

    for arrtype in ['SCI', 'ENG']:
        # create simulated dataset
        dataset = mocks.create_prescan_files(filedir=datadir, arrtype=arrtype,numfiles=1)

        # Generate bias with inflated values in the bad columns
        bias = np.full_like(dataset.all_data,bval)
        col_start = detector_areas[arrtype]['prescan_reliable']['r0c0'][1]
        bias[:,:,0:col_start] = bval * 5

        # Overwrite dataset with normal + exponential + bias
        rng = np.random.default_rng(seed)
        dataset.all_data[:,:,:] = (rng.normal(0, 1, size=dataset.all_data.shape)
                                   + rng.exponential(expmean, size=dataset.all_data.shape)
                                   - expmean # to keep DC contribution 0
                                   + bias)

        for return_full_frame in [True, False]:
            output_dataset = prescan_biassub(dataset, noise_maps, return_full_frame=return_full_frame)
            if np.any(np.abs(output_dataset[0].hdu_list['BIAS'].data - bval) > tol):
                raise Exception(f'Higher than expected error in bias measurement for hvon distribution.')

def test_bias_uniform_value():
    """Verify that function finds bias for uniform value."""

    tol = 1e-13
    bval = 1.

    ###### create simulated dataset
    datadir = os.path.join(os.path.dirname(__file__), "simdata")

    for arrtype in ['SCI', 'ENG']:
        # create simulated dataset
        dataset = mocks.create_prescan_files(filedir=datadir, arrtype=arrtype,numfiles=1)

        # Overwrite dataset with normal + exponential + bias
        dataset.all_data[:,:,:] = bval

        for return_full_frame in [True, False]:
            output_dataset = prescan_biassub(dataset, noise_maps=None, return_full_frame=return_full_frame)
            if np.max(np.abs(output_dataset.all_data)) > tol:
                raise Exception(f'Higher than expected error in bias measurement for uniform value.')

            if np.max(np.abs(output_dataset.all_err)) > tol:
                raise Exception(f'Higher than expected std. error in bias measurement for uniform value.')

def test_bias_offset():
    """Verify bias offset incorporated as expected"""

    # 10 counts higher than the bias in the image region.
    bias_offset = 10
    tol = 1e-13

    ###### create simulated dataset
    datadir = os.path.join(os.path.dirname(__file__), "simdata")

    for arrtype in ['SCI', 'ENG']:
        # create simulated dataset with 0 bias offset
        dataset_0 = mocks.create_prescan_files(filedir=datadir, arrtype=arrtype,numfiles=1)
        dataset_0.all_data[:,:,:] = 0.

        # create simulated dataset with 10 bias offset
        # bias_offset = 10 means the bias, as measured in the prescan, is
        # 10 counts higher than the bias in the image region.
        dataset_10 = dataset_0.copy()
        r0c0 = detector_areas[arrtype]['prescan']['r0c0']
        rows = detector_areas[arrtype]['prescan']['rows']
        cols = detector_areas[arrtype]['prescan']['cols']
        dataset_10.all_data[:,r0c0[0]:r0c0[0]+rows, r0c0[1]:r0c0[1]+cols] += bias_offset

        noise_maps0 = noise_maps.copy()
        noise_maps0.bias_offset = 0
        noise_maps10 = noise_maps.copy()
        noise_maps10.bias_offset = 10
        for return_full_frame in [True,False]:
            output_dataset_0 = prescan_biassub(dataset_0, noise_maps0, return_full_frame=return_full_frame)
            output_dataset_10 = prescan_biassub(dataset_10, noise_maps10, return_full_frame=return_full_frame)

            # Compare science image region only
            if return_full_frame:
                r0c0 = detector_areas[arrtype]['image']['r0c0']
                rows = detector_areas[arrtype]['image']['rows']
                cols = detector_areas[arrtype]['image']['cols']
                image_slice_0 = output_dataset_0.all_data[0,r0c0[0]:r0c0[0]+rows, r0c0[1]:r0c0[1]+cols]
                image_slice_10 = output_dataset_10.all_data[0,r0c0[0]:r0c0[0]+rows, r0c0[1]:r0c0[1]+cols]
            else:
                image_slice_0 = output_dataset_0.all_data[0]
                image_slice_10 = output_dataset_10.all_data[0]

            if not np.nanmax(np.abs(image_slice_0 - image_slice_10)) < tol:
                raise Exception(f"Bias offset subtraction did not produce the correct result. absmax value : {np.nanmax(np.abs(image_slice_0 - image_slice_10))}")


if __name__ == "__main__":
    test_prescan_sub()
    test_bias_zeros_frame()
    test_bias_hvoff()
    test_bias_hvon()
    test_bias_uniform_value()
    test_bias_offset()


import os
import pickle
import pytest
import corgidrp
import numpy as np
import corgidrp.mocks as mocks
from corgidrp.mocks import create_default_calibration_product_headers
from corgidrp.l2a_to_l2b import correct_bad_pixels
from corgidrp.data import Image, Dataset, BadPixelMap

old_err_tracking = corgidrp.track_individual_errors

data = np.ones([1024,1024])*2.
err = np.ones([1024,1024]) *0.5
dq = np.zeros([1024,1024], dtype = np.uint16)
prhd, exthd = create_default_calibration_product_headers()

def test_bad_pixels():

    corgidrp.track_individual_errors = False

    print("UT for pipeline step correct_bad_pixels")

    image1 = Image(data, pri_hdr = prhd, ext_hdr = exthd, err = err, dq = dq)
    dataset = Dataset([image1])

    corgidrp.track_individual_errors = old_err_tracking

    data_cube = dataset.all_data # 3D data cube for all frames in the dataset
    dq_cube = dataset.all_dq # 3D DQ array cube for all frames in the dataset
    # Add some CR
    col_cr_test = [12, 123, 234, 456, 678, 890]
    row_cr_test = [546, 789, 123, 43, 547, 675]
    for i_col in col_cr_test:
        for i_row in row_cr_test:
            dq_cube[0, i_col, i_row] = np.bitwise_or(dq_cube[0, i_col, i_row].astype(np.uint8), 128)

    history_msg = "Pixels affected by CR added"
    dataset.update_after_processing_step(history_msg, new_all_data=data_cube,
        new_all_dq=dq_cube)

    assert type(dataset) == corgidrp.data.Dataset

    # Generate bad pixel detector mask
    datadir = os.path.join(os.path.dirname(__file__), "testcalib")
    if not os.path.exists(datadir):
        os.mkdir(datadir)
    outputdir = os.path.join(os.path.dirname(__file__), "testcalib")
    if not os.path.exists(outputdir):
        os.mkdir(outputdir)
    col_bp_test=[12, 120, 234, 450, 678, 990]
    row_bp_test=[546, 89, 123, 243, 447, 675]
    bp_mask = mocks.create_badpixelmap_files(filedir=datadir,
        col_bp=col_bp_test, row_bp=row_bp_test)
    new_bp_mask = BadPixelMap(bp_mask.all_data[0], pri_hdr=bp_mask[0].pri_hdr.copy(),
                     ext_hdr=bp_mask[0].ext_hdr.copy(), input_dataset=bp_mask)

    assert type(new_bp_mask) == corgidrp.data.BadPixelMap
    
    # check the bpmap can be pickled (for CTC operations)
    pickled = pickle.dumps(new_bp_mask)
    pickled_bpmap = pickle.loads(pickled)
    assert np.all((new_bp_mask.data == pickled_bpmap.data))


    new_dataset = correct_bad_pixels(dataset, new_bp_mask)

    assert type(new_dataset) == corgidrp.data.Dataset

    # Checking that NaN are only at the expected locations
    nan_data = np.argwhere(np.isnan(new_dataset.all_data[0]))
    for ii in nan_data[:,0]:
        assert ii in col_cr_test or ii in col_bp_test
    for jj in nan_data[:,1]:
        assert jj in row_cr_test or jj in row_bp_test

    # Checking that bad pixels are only at the expected locations
    replaced_pix = np.argwhere(new_dataset.all_dq[0] != 0)
    for ii in replaced_pix[:,0]:
        assert ii in col_cr_test or ii in col_bp_test
    for jj in replaced_pix[:,1]:
        assert jj in row_cr_test or jj in row_bp_test

    # Checking that DQ bit for replaced pixels is set to 1 (Big-Endian)
    for ii in replaced_pix[:,0]:
        for jj in replaced_pix[:,1]:
            # Only bad pixels
            if new_dataset.all_dq[0][ii][jj] != 0:
                assert np.unpackbits(new_dataset.all_dq[0][ii][jj].astype('uint8'))[6] == 1

    # Checking that bad pixels are at the expected locations
    bp_dq = np.where(new_dataset.all_dq[0] == 4)
    for ii in bp_dq[0]:
        assert ii in col_bp_test
    for jj in bp_dq[1]:
        assert jj in row_bp_test

    # Checking that CR are at the expected locations
    cr_dq = np.where(new_dataset.all_dq[0] == 128)
    for ii in cr_dq[0]:
        assert ii in col_cr_test
    for jj in cr_dq[1]:
        assert jj in row_cr_test

    # Checking that coincident bad pixels and CR are at the expected locations
    bp_cr_dq = np.where(new_dataset.all_dq[0] == 132)
    for ii in bp_cr_dq[0]:
        assert ii in col_bp_test and ii in col_cr_test
    for jj in bp_cr_dq[1]:
        assert jj in row_bp_test and jj in row_cr_test


    # save and reload bad pixel map
    new_bp_mask.save(filedir=outputdir, filename="sim_bp_map_cal.fits")
    new_bp_mask_2 = BadPixelMap(os.path.join(outputdir, "sim_bp_map_cal.fits"))
    
    # check the bpmap can be pickled (for CTC operations)
    pickled = pickle.dumps(new_bp_mask_2)
    pickled_bpmap = pickle.loads(pickled)
    assert np.all((new_bp_mask_2.data == pickled_bpmap.data))

    print("UT passed")

if __name__ == '__main__':
    test_bad_pixels()



import pytest
import numpy as np
from corgidrp.mocks import create_default_L1_headers
from corgidrp.data import Image, Dataset
import corgidrp.l2a_to_l2b as l2a_to_l2b

def test_emgain_div():
    data = np.ones([1024,1024]) * 2 
    err = np.ones([1,1024,1024]) *0.5
    prhd, exthd = create_default_L1_headers()
    exthd["EMGAIN_C"] = 1000
    image1 = Image(data,pri_hdr = prhd, ext_hdr = exthd, err = err)
    image2 = image1.copy()
    dataset=Dataset([image1, image2])
    
    ###### perform em_gain division
    gain_dataset = l2a_to_l2b.em_gain_division(dataset)
    assert("same" in str(gain_dataset[0].ext_hdr["HISTORY"]))

    emgain = gain_dataset[0].ext_hdr['EMGAIN_C']
    
    # check the level of the dataset
    assert np.mean(gain_dataset.all_data) == pytest.approx(np.mean(dataset.all_data)/emgain, abs=1e-3)
    assert np.mean(gain_dataset.all_err) == pytest.approx(np.mean(dataset.all_err)/emgain, abs=1e-3)
    assert gain_dataset[0].ext_hdr["BUNIT"] == "detected electrons"
    assert gain_dataset[0].err_hdr["BUNIT"] == "detected electrons"
    
    # check non-unique emgain
    emgain1 = 100
    dataset[1].ext_hdr['EMGAIN_C'] = emgain1
    gain_dataset = l2a_to_l2b.em_gain_division(dataset)
    assert("different" in str(gain_dataset[0].ext_hdr["HISTORY"]))
    assert np.mean(gain_dataset.all_data[0]) == pytest.approx(np.mean(dataset.all_data[0])/emgain, abs=1e-3)
    assert np.mean(gain_dataset.all_data[1]) == pytest.approx(np.mean(dataset.all_data[1])/emgain1, abs=1e-3)
    assert np.mean(gain_dataset.all_err[0]) == pytest.approx(np.mean(dataset.all_err[0])/emgain, abs=1e-3)
    assert np.mean(gain_dataset.all_err[1]) == pytest.approx(np.mean(dataset.all_err[1])/emgain1, abs=1e-3)
    

if __name__ == "__main__":
    test_emgain_div()


"""
Unit test suite for the calibrate_kgain module.

Frames for unit tests are simulated SCI-size frames made with Gaussian and Poisson 
noises included. The assumed flux map is a realistic pupil image made from 
TVAC frames.
"""

import os
import pytest
import warnings
import numpy as np
import pandas as pd
from pathlib import Path

import test_check
from corgidrp import check
from corgidrp.data import Image, Dataset
from corgidrp.mocks import (make_fluxmap_image, nonlin_coefs)
from corgidrp.calibrate_kgain import (calibrate_kgain, CalKgainException, kgain_params_default)



######################## function definitions ###############################

def count_contiguous_repeats(arr):
    """
    This function returns the count of the contiguous repeated exposure times.
    If input array is empty, returns empty array.

    Args:
        arr (np.array): 1d array of exposure times.

    Returns:
        counts (list): count of the contiguous repeated exposure times
    """

    if isinstance(arr, (np.ndarray, list)) and len(arr) == 0:
        return []
    counts = []
    current_count = 1
    for i in range(1, len(arr)):
        if arr[i] == arr[i - 1]:
            current_count += 1
        else:
            counts.append(current_count)
            current_count = 1
    counts.append(current_count)  # append the count of the last element
    return counts

def setup_module():
    """
    Sets up module
    """
    global n_cal, n_mean, kgain_in
    global dataset_kg
    ############### make stacks with simulated frames ##################
    np.random.seed(8585)

    # path to nonlin table made from running calibrate_nonlin.py on TVAC frames
    # table used only to choose parameters to make analytic nonlin functions
    here = os.path.abspath(os.path.dirname(__file__))
    nonlin_table_path = Path(here,'test_data','nonlin_table_TVAC.txt')
    nonlin_flag = False # True adds nonlinearity to simulated frames

    # Load the arrays needed for calibrate_nonlin function from the .npz file
    loaded = np.load(Path(here,'test_data','nonlin_arrays_ut.npz'))
    # Access the arrays needed for calibrate_nonlin function
    exp_time_stack_arr0 = loaded['array1']
    time_stack_arr0 = loaded['array2']
    len_list0 = loaded['array3']
    # Reducing the number of frames used in unit tests (each has 5 substacks).
    # Set to False to run the same test as in the IIT pipeline
    rn_in = 130 # read noise in e-
    if True:
        n_cal = 3
        exp_time_stack_arr0 = np.delete(exp_time_stack_arr0, np.s_[n_cal*5:])
        time_stack_arr0 = np.delete(time_stack_arr0, np.s_[n_cal*5:])
        # Update len_list0
        len_list0[0] = n_cal*5
        # Usual number of frames to deal with real rn values is ~200
        rn_in = 130/np.sqrt(200/n_cal) # read noise in e-

    # Load the flux map
    fluxmap_init =  np.load(Path(here,'test_data','FluxMap1024.npy'))
    fluxmap_init[fluxmap_init < 50] = 0 # cleanup flux map a bit
    fluxMap = 0.8*fluxmap_init # e/s/px, for G = 1

    # assumed detector parameters
    kgain_in = 8.7 # e-/DN
    bias = 2000 # e-
    actual_gain = 1.0

    # cubic function nonlinearity for emgain of 1
    if nonlin_flag:
        coeffs_1, DNs, _ = nonlin_coefs(nonlin_table_path,1.0,3)
    else:
        coeffs_1 = [0.0, 0.0, 0.0, 1.0]
        _, DNs, _ = nonlin_coefs(nonlin_table_path,1.0,3)

    frame_list = []
    # make some uniform frames with emgain = 1 (must be unity) P.S. IIT would use ~30
    n_mean = 3
    for j in range(n_mean):
        image_sim = make_fluxmap_image(fluxMap,bias,kgain_in,rn_in, 1, 7.0,coeffs_1,
            nonlin_flag=nonlin_flag)
        # Datetime cannot be duplicated
        image_sim.ext_hdr['DATETIME'] = time_stack_arr0[j]
        # Temporary keyword value. Mean frame is TBD
        image_sim.pri_hdr['OBSNAME'] = 'MNFRAME'
        frame_list.append(image_sim)

    index = 0
    iG = 0 # doing only the em gain = 1 case
    g = actual_gain # Note: Same value for all frames used to calibrate K-gain
    exp_time_loop = exp_time_stack_arr0[index:index+len_list0[iG]]
    index = index + len_list0[iG]
    if nonlin_flag:
        coeffs, _, vals = nonlin_coefs(nonlin_table_path,g,3)
    else:
        coeffs = [0.0, 0.0, 0.0, 1.0]
        vals = np.ones(len(DNs))

    exp_repeat_counts = count_contiguous_repeats(exp_time_loop)
    for j in range(len(exp_repeat_counts)):
        for t in range(exp_repeat_counts[j]):
            # Simulate full frame
            exp_time = exp_time_loop[t+j*exp_repeat_counts[j]]
            image_sim = make_fluxmap_image(fluxMap,bias,kgain_in,rn_in,g,
                                exp_time,coeffs,nonlin_flag=nonlin_flag,
                                divide_em=True)
            image_sim.ext_hdr['DATETIME'] = time_stack_arr0[t+j*exp_repeat_counts[j]]
            # OBSNAME has no KGAIN value, but NONLIN
            image_sim.pri_hdr['OBSNAME'] = 'NONLIN'
            frame_list.append(image_sim)
    dataset_kg = Dataset(frame_list)


def teardown_module():
    """
    Run at end of tests. Deletes variables
    """
    global n_cal, n_mean, kgain_in
    global dataset_kg

    del n_cal, n_mean, kgain_in
    del dataset_kg

# set input parameters for calibrate_kgain function
min_val = 800
max_val = 3000
binwidth = 68

def test_expected_results_sub():
    """Outputs are as expected, for imported frames."""
    kgain = calibrate_kgain(dataset_kg, n_cal, n_mean, min_val, max_val, binwidth)
        
    signal_bins_N = kgain_params_default['signal_bins_N']
    # kgain - should be close to the assumed value
    assert np.isclose(round(kgain.value,1), kgain_in, atol=0.5)
    assert np.all(np.equal(kgain.ptc.shape, (signal_bins_N,2)))

    # test bad input for kgain_params
    kgain_params_bad = kgain_params_default.copy()
    kgain_params_bad['colroi2'] = 'foo'
    with pytest.raises(TypeError):
        calibrate_kgain(dataset_kg, n_cal, n_mean, min_val, max_val, binwidth,
                        kgain_params=kgain_params_bad)

def test_psi():
    """These three below must be positive scalar integers."""
    check_list = test_check.psilist
    # min_val
    for perr in check_list:
        with pytest.raises(TypeError):
            calibrate_kgain(dataset_kg, n_cal, n_mean, perr, max_val, binwidth)
    # max_val
    for perr in check_list:
        with pytest.raises(TypeError):
            calibrate_kgain(dataset_kg, n_cal, n_mean, min_val, perr, binwidth)

    # binwidth
    for perr in check_list:
        with pytest.raises(TypeError):
            calibrate_kgain(dataset_kg, n_cal, n_mean, min_val, max_val, perr)
      
def test_binwidth():
    """binwidth must be >= 10."""
    with pytest.raises(CalKgainException):
        calibrate_kgain(dataset_kg, n_cal, n_mean, min_val, max_val, 9)
 
if __name__ == '__main__':
    print('Running test_expected_results_sub')
    test_expected_results_sub()
    print('Running test_psi')
    test_psi()
    print('Running test_binwidth')
    test_binwidth()



import os
import glob
import pytest
import numpy as np
import corgidrp
import corgidrp.data as data
import corgidrp.mocks as mocks
import corgidrp.detector as detector
import corgidrp.flat as flat
import corgidrp.l2a_to_l2b as l2a_to_l2b

old_err_tracking = corgidrp.track_individual_errors

np.random.seed(9292)
def test_flat_div():
    """
    Generate mock input data and pass into flat division function
    """
    corgidrp.track_individual_errors = True # this test uses individual error components

    ###### create simulated data
    # check that simulated data folder exists, and create if not
    datadir = os.path.join(os.path.dirname(__file__), "simdata")
    if not os.path.exists(datadir):
        os.mkdir(datadir) 
    mocks.create_simflat_dataset(filedir=datadir)
    
    # simulated images to be checked in flat division
    simdata_filenames=glob.glob(os.path.join(datadir, "sim_flat*.fits"))
    simflat_dataset=data.Dataset(simdata_filenames)
     
    # creat one dummy flat field perform flat division
    mocks.create_flatfield_dummy(filedir=datadir)
	#test data architecture
    flat_filenames = glob.glob(os.path.join(datadir, "flat_field*.fits"))
    flat_dataset = data.Dataset(flat_filenames)
    #check that data is consistently modified
    flat_dataset.all_data[0,0,0] = 1
    assert flat_dataset[0].data[0,0] == 1
    flat_dataset[0].data[0,0] = 1
    assert flat_dataset.all_data[0,0,0] == 1
    
    ###### create flatfield
    flat_frame = flat.create_flatfield(flat_dataset)
    # check the level of counts in flatfield is approximately correct
    assert np.mean(flat_frame.data) == pytest.approx(1, abs=1e-2)
    # check that the error is determined correctly
    assert np.array_equal(np.std(flat_dataset.all_data, axis = 0)/np.sqrt(len(flat_dataset)), flat_frame.err[0])
	# save flatfield
    calibdir = os.path.join(os.path.dirname(__file__), "testcalib")
    flat_filename = "sim_flat_calib.fits"
    if not os.path.exists(calibdir):
        os.mkdir(calibdir)
    flat_frame.save(filedir=calibdir, filename=flat_filename)
    
    ###### perform flat division
    # load in the flatfield
    flat_filepath = os.path.join(calibdir, flat_filename)
    flatfield = data.FlatField(flat_filepath)
    # divide the simulated dataset with the dummy flatfield
    flatdivided_dataset = l2a_to_l2b.flat_division(simflat_dataset, flatfield)
    
    # perform checks after the flat divison
    assert(flat_filename in str(flatdivided_dataset[0].ext_hdr["HISTORY"]))
    # check the level of the dataset is now approximately 100
    assert np.mean(flatdivided_dataset.all_data) == pytest.approx(150, abs=2e-2)
    # check the propagated errors
    assert flatdivided_dataset[0].err_hdr["Layer_2"] == "FlatField_error"
    print("mean of all simulated data",np.mean(simflat_dataset.all_data))
    print("mean of all simulated data error",np.mean(simflat_dataset.all_err) )
    print("mean of all flat divided data:", np.mean(flatdivided_dataset.all_data))
    print("mean of all flat divided data errors:", np.mean(flatdivided_dataset.all_err))
    print("mean of flatfield:", np.mean(flatfield.data))
    print("mean of flatfield err:", np.mean(flatfield.err))
    
    err_flatdiv=np.mean(flatdivided_dataset.all_err)
    err_estimated=np.sqrt(((np.mean(flatfield.data))**2)*(np.mean(simflat_dataset.all_err))**2+((np.mean(simflat_dataset.all_data))**2)*(np.mean(flatfield.err))**2)
    print("mean of all flat divided data errors:",err_flatdiv)
    print("Error estimated:",err_estimated)
    assert(err_flatdiv == pytest.approx(err_estimated, abs = 1e-2))
    
    # print(flatdivided_dataset[0].ext_hdr)
    corgidrp.track_individual_errors = old_err_tracking
    

    ### Check to make sure DQ gets set when flatfield zero. ###

    ## Injected some 0s. 
    pixel_list = [[110,120],[50,50], [100,100]]
    for pixel in pixel_list:
        flatfield.data[pixel[0],pixel[1]] = 0.
    
    ## Perform flat division
    flatdivided_dataset_w_zeros = l2a_to_l2b.flat_division(simflat_dataset, flatfield)

    ## Check that all the pixels that were zeroed out have the DQ flag set to 4. 
    for pixel in pixel_list:
        for i in range(len(simdata_filenames)):
            assert np.bitwise_and(flatdivided_dataset_w_zeros.all_dq[i,pixel[0],pixel[1]],4)

    
if __name__ == "__main__":
    test_flat_div()


# A file that holds the functions that transmogrify l2a data to l2b data
import numpy as np
import copy
import corgidrp.data as data
from corgidrp.darks import build_synthesized_dark
from corgidrp.detector import detector_areas

def add_photon_noise(input_dataset):
    """
    Propagate the photon/shot noise determined from the image signal to the error map. 

    Args:
       input_dataset (corgidrp.data.Dataset): a dataset of Images (L2a-level)

    Returns:
        corgidrp.data.Dataset: photon noise propagated to the image error extensions of the input dataset
    """
    # you should make a copy the dataset to start
    phot_noise_dataset = input_dataset.copy() # necessary at all?

    for i, frame in enumerate(phot_noise_dataset.frames):
        try: # use measured gain if available TODO change hdr name if necessary
            em_gain = phot_noise_dataset[i].ext_hdr["EMGAIN_M"]
        except:
            em_gain = frame.ext_hdr.get("EMGAIN_A", 0)
            if em_gain > 0: # use EM applied EM gain if available
                em_gain = em_gain
            else: # otherwise use commanded EM gain
                em_gain = frame.ext_hdr.get("EMGAIN_C", 0)
        phot_err = np.sqrt(frame.data)
        #add excess noise in case of em_gain
        if em_gain > 1:
            phot_err *= np.sqrt(2)           
        frame.add_error_term(phot_err, "photnoise_error")
    
    new_all_err = np.array([frame.err for frame in phot_noise_dataset.frames])        
    
    history_msg = "photon noise propagated to error map"
    # update the output dataset
    phot_noise_dataset.update_after_processing_step(history_msg, new_all_err = new_all_err)

    return phot_noise_dataset


def dark_subtraction(input_dataset, dark, detector_regions=None, outputdir=None):
    """

    Perform dark subtraction of a dataset using the corresponding dark frame.  The dark frame can be either a synthesized master dark (made for any given EM gain and exposure time)
    or for a traditional master dark (average of darks taken at the EM gain and exposure time of the corresponding observation).  The master dark is also saved if it is of the synthesized type after it is built.

    Args:
        input_dataset (corgidrp.data.Dataset): a dataset of Images that need dark subtraction (L2a-level)
        dark (corgidrp.data.Dark or corgidrp.data.DetectorNoiseMaps): If dark is of the corgidrp.data.Dark type, dark subtraction will be done immediately.
            If dark is of the corgidrp.data.DetectorNoiseMaps type, a synthesized master is created using calibrated noise maps for the EM gain and exposure time used in the frames in input_dataset.
        detector_regions: (dict):  A dictionary of detector geometry properties.  Keys should be as found in detector_areas in detector.py. Defaults to detector_areas in detector.py.
        outputdir (string): Filepath for output directory where to save the master dark if it is a synthesized master dark.  Defaults to current directory.

    Returns:
        corgidrp.data.Dataset: a dark-subtracted version of the input dataset including error propagation
    """
    _, unique_vals = input_dataset.split_dataset(exthdr_keywords=['EXPTIME', 'EMGAIN_C', 'KGAINPAR'])
    if len(unique_vals) > 1:
        raise Exception('Input dataset should contain frames of the same exposure time, commanded EM gain, and k gain.')

    if detector_regions is None:
        detector_regions = detector_areas
    # you should make a copy the dataset to start
    darksub_dataset = input_dataset.copy()
    rows = detector_regions['SCI']['frame_rows']
    cols = detector_regions['SCI']['frame_cols']
    im_rows = detector_regions['SCI']['image']['rows']
    im_cols = detector_regions['SCI']['image']['cols']

    if type(dark) is data.DetectorNoiseMaps:
        if input_dataset.frames[0].data.shape == (rows, cols):
            full_frame = True
        elif input_dataset.frames[0].data.shape == (im_rows, im_cols):
            full_frame = False
        else:
            raise Exception('Frames in input_dataset do not have valid SCI full-frame or image dimensions.')
        dark = build_synthesized_dark(input_dataset, dark, detector_regions=detector_regions, full_frame=full_frame)
        if outputdir is None:
            outputdir = '.' #current directory
        dark.save(filedir=outputdir)
    elif type(dark) is data.Dark:
        if 'PC_STAT' in dark.ext_hdr:
            if dark.ext_hdr['PC_STAT'] != 'analog master dark':
                raise Exception('The input \'dark\' is a photon-counted dark and cannot be used in the dark_subtraction step function, which is intended for analog frames.')
        # In this case, the Dark loaded in should already match the arry dimensions
        # of input_dataset, specified by full_frame argument of build_trad_dark
        # when this Dark was built
        if (dark.ext_hdr['EXPTIME'], dark.ext_hdr['EMGAIN_C'], dark.ext_hdr['KGAINPAR']) != unique_vals[0]:
            raise Exception('Dark should have the same EXPTIME, EMGAIN_C, and KGAINPAR as input_dataset.')
    else:
        raise Exception('dark type should be either corgidrp.data.Dark or corgidrp.data.DetectorNoiseMaps.')

    darksub_cube = darksub_dataset.all_data - dark.data

    # propagate the error of the dark frame
    if hasattr(dark, "err"):
        darksub_dataset.add_error_term(dark.err[0], "dark_error")
    else:
        raise Warning("no error attribute in the dark frame")

    if hasattr(dark, "dq"):
        new_all_dq = np.bitwise_or(darksub_dataset.all_dq, dark.dq)
    else:
        new_all_dq = None

    #darksub_dataset.all_err = np.array([frame.err for frame in darksub_dataset.frames])
    history_msg = "Dark subtracted using dark {0}.  Units changed from detected electrons to photoelectrons.".format(dark.filename)

    # update the output dataset with this new dark subtracted data and update the history
    darksub_dataset.update_after_processing_step(history_msg, new_all_data=darksub_cube, new_all_dq = new_all_dq, header_entries = {"BUNIT":"Photoelectrons"})

    return darksub_dataset

def flat_division(input_dataset, flat_field):
    """

    Divide the dataset by the master flat field.

    Args:
        input_dataset (corgidrp.data.Dataset): a dataset of Images (L2a-level)
        flat_field (corgidrp.data.FlatField): a master flat field to divide by

    Returns:
        corgidrp.data.Dataset: a version of the input dataset with the flat field divided out
    """

     # copy of the dataset
    flatdiv_dataset = input_dataset.copy()

    #Divide by the master flat
    flatdiv_cube = flatdiv_dataset.all_data /  flat_field.data

    #Find where the flat_field is 0 and set a DQ flag: 
    where_zero = np.where(flat_field.data == 0)
    flatdiv_dq = copy.deepcopy(flatdiv_dataset.all_dq)
    for i in range(len(flatdiv_dataset)):
       flatdiv_dq[i][where_zero] = np.bitwise_or(flatdiv_dataset[i].dq[where_zero], 4)

    # propagate the error of the master flat frame
    if hasattr(flat_field, "err"):
        flatdiv_dataset.rescale_error(1/flat_field.data, "FlatField")
        flatdiv_dataset.add_error_term(flatdiv_dataset.all_data*flat_field.err[0]/(flat_field.data**2), "FlatField_error")
    else:
        raise Warning("no error attribute in the FlatField")

    history_msg = "Flat calibration done using Flat field {0}".format(flat_field.filename)

    # update the output dataset with this new flat calibrated data and update the history
    flatdiv_dataset.update_after_processing_step(history_msg,new_all_data=flatdiv_cube, new_all_dq = flatdiv_dq)

    return flatdiv_dataset

def frame_select(input_dataset, bpix_frac=1., allowed_bpix=0, overexp=False, tt_rms_thres=None, tt_bias_thres=None, discard_bad=True):
    """

    Selects the frames that we want to use for further processing.

    Args:
        input_dataset (corgidrp.data.Dataset): a dataset of Images (L2a-level)
        bpix_frac (float): greater than fraction of the image needs to be bad to discard. Default: 1.0 (not used)
        allowed_bpix (int): sum of DQ values that are allowed and not counted towards to bpix fraction
                            (e.g., 6 means 2 and 4 are not considered bad).
                            Default is 0 (all nonzero DQ flags are considered bad)
        overexp (bool): if True, removes frames where the OVEREXP keyword is True. Default: False
        tt_rms_thres (float): maximum allowed RMS tip or tilt in image to be considered good. Default: None (not used)
        tt_bias_thres (float): maximum allowed bias in tip/tilt over the course of an image to be consdiered good. Default: None (not used)
        discard_bad (bool): if True, drops the bad frames rather than keeping them through processing
        
    Returns:
        corgidrp.data.Dataset: a version of the input dataset with only the frames we want to use
    """
    pruned_dataset = input_dataset.copy()
    reject_flags = np.zeros(len(input_dataset))
    reject_reasons = {}

    disallowed_bits = np.invert(allowed_bpix) # invert the mask

    for i, frame in enumerate(pruned_dataset.frames):
        reject_reasons[i] = [] # list of rejection reasons
        if bpix_frac < 1:
            masked_dq = np.bitwise_and(frame.dq, disallowed_bits) # handle allowed_bpix values
            numbadpix = np.size(np.where(masked_dq > 0)[0])
            frame_badpix_frac = numbadpix / np.size(masked_dq)
            # if fraction of bad pixel over threshold, mark is as bad
            if frame_badpix_frac > bpix_frac:
                reject_flags[i] += 1
                reject_reasons[i].append("bad pix frac {0:.5f} > {1:.5f}"
                                         .format(frame_badpix_frac, bpix_frac))
        frame.ext_hdr['FRMSEL01'] = (bpix_frac, "Bad Pixel Fraction < This Value. Doesn't include DQflags summed to {0}".format(allowed_bpix)) # record selection criteria

        if overexp:
            if frame.ext_hdr['OVEREXP']:
                reject_flags[i] += 2 # use distinct bits in case it's useful
                reject_reasons[i].append("OVEREXP = T")
        frame.ext_hdr['FRMSEL02'] = (overexp, "Are we selecting on the OVEREXP flag?") # record selection criteria

        if tt_rms_thres is not None:
            if frame.ext_hdr['Z2VAR'] > tt_rms_thres:
                reject_flags[i] += 4 # use distinct bits in case it's useful
                reject_reasons[i].append("tip rms (Z2VAR) {0:.1f} > {1:.1f}"
                                         .format(frame.ext_hdr['Z2VAR'], tt_rms_thres))
            if frame.ext_hdr['Z3VAR'] > tt_rms_thres:
                reject_flags[i] += 8 # use distinct bits in case it's useful
                reject_reasons[i].append("tilt rms (Z3VAR) {0:.1f} > {1:.1f}"
                                         .format(frame.ext_hdr['Z3VAR'], tt_rms_thres))
        frame.ext_hdr['FRMSEL03'] = (tt_rms_thres, "tip rms (Z2VAR) threshold") # record selection criteria
        frame.ext_hdr['FRMSEL04'] = (tt_rms_thres, "tilt rms (Z3VAR) threshold") # record selection criteria
        
        if tt_bias_thres is not None:
            if frame.ext_hdr['Z2RES'] > tt_bias_thres:
                reject_flags[i] += 16 # use distinct bits in case it's useful
                reject_reasons[i].append("tip bias (Z2RES) {0:.1f} > {1:.1f}"
                                         .format(frame.ext_hdr['Z2RES'], tt_bias_thres))
            if frame.ext_hdr['Z3RES'] > tt_bias_thres:
                reject_flags[i] += 32 # use distinct bits in case it's useful
                reject_reasons[i].append("tilt bias (Z3RES) {0:.1f} > {1:.1f}"
                                         .format(frame.ext_hdr['Z3RES'], tt_bias_thres))
        frame.ext_hdr['FRMSEL05'] = (tt_bias_thres, "tip bias (Z2RES) threshold") # record selection criteria
        frame.ext_hdr['FRMSEL06'] = (tt_bias_thres, "tilt bias (Z3RES) threshold") # record selection criteria
                
        # if rejected, mark as bad in the header
        if reject_flags[i] > 0:
            frame.ext_hdr['IS_BAD'] = True
                                
    good_frames = np.where(reject_flags == 0)
    bad_frames = np.where(reject_flags > 0)
    # check that we didn't remove all of the good frames
    if np.size(good_frames) == 0:
        raise ValueError("No good frames were selected. Unable to continue")

    # if we need to discard bad, do that here. 
    if discard_bad:
        pruned_frames = pruned_dataset.frames[good_frames]
        pruned_dataset = data.Dataset(pruned_frames)
        
        # history message of which frames were removed and why
        history_msg = "Removed {0} frames as bad:".format(np.size(bad_frames))
    else:
        # history message of which frames were marked and why
        history_msg = "Marked {0} frames as bad:".format(np.size(bad_frames))

    for bad_index in bad_frames[0]:
        bad_frame = input_dataset.frames[bad_index]
        bad_reasons = "; ".join(reject_reasons[bad_index])
        history_msg += " {0} ({1}),".format(bad_frame.filename, bad_reasons)
    history_msg = history_msg[:-1] # remove last comma or :

    pruned_dataset.update_after_processing_step(history_msg)

    return pruned_dataset

def convert_to_electrons(input_dataset, k_gain):
    """

    Convert the data from ADU to electrons.
    TODO: Establish the interaction with the CalDB for obtaining gain calibration

    Args:
        input_dataset (corgidrp.data.Dataset): a dataset of Images (L2a-level)
        k_gain (corgidrp.data.KGain): KGain calibration file

    Returns:
        corgidrp.data.Dataset: a version of the input dataset with the data in electrons
    """
   # you should make a copy the dataset to start
    kgain_dataset = input_dataset.copy()
    kgain_cube = kgain_dataset.all_data

    kgain = k_gain.value #extract from caldb
    kgainerr = k_gain.error[0]
    # x = c*kgain, where c (counts beforehand) and kgain both have error, so do propogation of error due to the product of 2 independent sources
    kgain_error = np.zeros_like(input_dataset.all_err)
    kgain_tot_err = (np.sqrt((kgain*kgain_dataset.all_err[:,0,:,:])**2 + (kgain_cube*kgainerr)**2))
    kgain_error[:,0,:,:] = kgain_tot_err
    kgain_cube *= kgain
    
    history_msg = "data converted to detected EM electrons by kgain {0}".format(str(kgain))

    # update the output dataset with this converted data and update the history
    kgain_dataset.update_after_processing_step(history_msg, new_all_data=kgain_cube, new_all_err=kgain_error, header_entries = {"BUNIT":"detected EM electrons", "KGAINPAR":kgain, 
                                                                                    "KGAIN_ER": k_gain.error[0], "RN":k_gain.ext_hdr['RN'], "RN_ERR":k_gain.ext_hdr["RN_ERR"]})
    return kgain_dataset

def em_gain_division(input_dataset):
    """

    Convert the data from detected EM electrons to detected electrons by dividing by the EM gain.
    Update the change in units in the header [detected electrons].

    Args:
        input_dataset (corgidrp.data.Dataset): a dataset of Images (L2a-level)

    Returns:
        corgidrp.data.Dataset: a version of the input dataset with the data in units "detected electrons"
    """

    # you should make a copy the dataset to start
    emgain_dataset = input_dataset.copy()
    emgain_cube = emgain_dataset.all_data
    emgain_error = emgain_dataset.all_err

    for i in range(len(emgain_dataset)):
        try: # use measured gain if available TODO change hdr name if necessary
            emgain = emgain_dataset[i].ext_hdr["EMGAIN_M"]
        except:
            emgain = emgain_dataset[i].ext_hdr["EMGAIN_A"]
            if emgain > 0: # use EM applied EM gain if available
                emgain = emgain_dataset[i].ext_hdr["EMGAIN_A"]
            else: # otherwise use commanded EM gain
                emgain = emgain_dataset[i].ext_hdr["EMGAIN_C"]
        emgain_cube[i] /= emgain
        emgain_error[i] /= emgain

    dataset_list, _ = emgain_dataset.split_dataset(exthdr_keywords=['EMGAIN_C'])
    if len(dataset_list) > 1:
        history_msg = "data divided by EM gain for dataset with frames with different commanded EM gains"
    else:
        history_msg = "data divided by EM gain for dataset with frames with the same commanded EM gain"

    # update the output dataset with this EM gain divided data and update the history
    emgain_dataset.update_after_processing_step(history_msg, new_all_data=emgain_cube, new_all_err=emgain_error, header_entries = {"BUNIT":"detected electrons"})

    return emgain_dataset

def cti_correction(input_dataset, pump_trap_cal):
    """

    Apply the CTI correction to the dataset.

    Currently a no-op step function. 

    Args:
        input_dataset (corgidrp.data.Dataset): a dataset of Images (L2a-level)
        pump_trap_cal (corgidrp.data.TrapCalibration): Pump trap calibration file

    Returns:
        corgidrp.data.Dataset: a version of the input dataset with the CTI correction applied
    """
    # also remember to update CTI_CORR ext header keyword
    return input_dataset.copy()


def correct_bad_pixels(input_dataset, bp_mask):

    """

    Correct for bad pixels: Bad pixels are identified as part of the data
        calibration. This function replaces bad pixels by NaN values. It also
        updates its DQ storing the type of bad pixel at each bad pixel location,
        and it records the fact that the pixel has been replaced by NaN.

    Args:
        input_dataset (corgidrp.data.Dataset): a dataset of Images (L2a-level)
        bp_mask (corgidrp.data.BadPixelMap): Bad-pixel mask built from the bad
        pixel calibration file.


    Returns:
        corgidrp.data.Dataset: a version of the input dataset with bad detector
        pixels and cosmic rays replaced by NaNs

    """

    data = input_dataset.copy()
    data_cube = data.all_data
    dq_cube = data.all_dq.astype(np.uint8)

    for i in range(data_cube.shape[0]):
        # combine DQ and BP masks
        bp_dq_mask = np.bitwise_or(dq_cube[i],bp_mask.data.astype(np.uint8))
        # mask affected pixels with NaN
        bp = np.where(bp_dq_mask != 0)
        data_cube[i, bp[0], bp[1]] = np.nan
        # Update DQ to keep track of replaced bad pixel values
        bp_dq_mask[bp[0], bp[1]]=np.bitwise_or(bp_dq_mask[bp[0], bp[1]], 2)
        dq_cube[i] = bp_dq_mask

    history_msg = "removed pixels affected by bad pixels"
    data.update_after_processing_step(history_msg, new_all_data=data_cube,
        new_all_dq=dq_cube)

    return data

def desmear(input_dataset, detector_params):
    """
    EXCAM has no shutter, and so continues to illuminate the detector during
    readout. This creates a "smearing" effect into the resulting images. The
    desmear function corrects for this effect. There are a small number of use
    cases for not desmearing data (e.g. time-varying raster data).

    Args:
        input_dataset (corgidrp.data.Dataset): a dataset of Images (L2a-level)
        detector_params (corgidrp.data.DetectorParams): a calibration file storing detector calibration values

    Returns:
        corgidrp.data.Dataset: a version of the input dataset with desmear applied

    """

    data = input_dataset.copy()
    data_cube = data.all_data

    rowreadtime_sec = detector_params.params['ROWREADT']

    for i in range(data_cube.shape[0]):
        exptime_sec = float(data[i].ext_hdr['EXPTIME'])
        smear = np.zeros_like(data_cube[i])
        m = len(smear)
        for r in range(m):
            columnsum = 0
            for s in range(r+1):
                columnsum = columnsum + rowreadtime_sec/exptime_sec*((1
                + rowreadtime_sec/exptime_sec)**((s+1)-(r+1)-1))*data_cube[i,s,:]
            smear[r,:] = columnsum
        data_cube[i] -= smear

    history_msg = "Desmear applied to data"
    header_update = {'DESMEAR' : True}
    data.update_after_processing_step(history_msg, new_all_data=data_cube, header_entries=header_update)

    return data

def update_to_l2b(input_dataset):
    """
    Updates the data level to L2b. Only works on L2a data.

    Currently only checks that data is at the L2a level

    Args:
        input_dataset (corgidrp.data.Dataset): a dataset of Images (L2a-level)

    Returns:
        corgidrp.data.Dataset: same dataset now at L2b-level
    """
    # check that we are running this on L1 data
    for orig_frame in input_dataset:
        if orig_frame.ext_hdr['DATALVL'] != "L2a":
            err_msg = "{0} needs to be L2a data, but it is {1} data instead".format(orig_frame.filename, orig_frame.ext_hdr['DATALVL'])
            raise ValueError(err_msg)

    # we aren't altering the data
    updated_dataset = input_dataset.copy(copy_data=False)

    for frame in updated_dataset:
        # update header
        frame.ext_hdr['DATALVL'] = "L2b"
        # update filename convention. The file convention should be
        # "CGI_[dataleel_*]" so we should be same just replacing the just instance of L1
        frame.filename = frame.filename.replace("_L2a", "_L2b", 1)

    history_msg = "Updated Data Level to L2b"
    updated_dataset.update_after_processing_step(history_msg)

    return updated_dataset


import numpy as np
from corgidrp.data import BadPixelMap, Dataset, DetectorNoiseMaps
from corgidrp.darks import build_synthesized_dark

def create_bad_pixel_map(dataset, master_dark, master_flat, dthresh = 5., ffrac = 0.8, fwidth = 32, dark_outputdir=None):
    """
    Compute a fixed bad pixel map for EXCAM from a master dark and flat.


    Args:
        dataset (corgidrp.data.Dataset): A dataset that is ignored by this class. 
        master_dark (corgidrp.data.Dark or corgidrp.data.DetectorNoiseMaps): 
            a dark frame, or, if it is a Noise Map,  a synthesized master is created using 
            calibrated noise maps for the EM gain and exposure time used in master_flat        
        master_flat (corgidrp.data.FlatField): A master flat field object.
        dthresh (float): Number of standard deviations above the mean to threshold for hot pixels. Must be >= 0.
        ffrac (float): Fraction of local mean value below which poorly-functioning pixels are flagged. Must be >=0 and < 1.
        fwidth (int): Number of pixels to include in local mean check with ffrac. Must be >0.
        dark_outputdir (str): if not None, a file directory to save any synthetic darks created from noise maps

    Returns:
        corgi.data.BadPixelMap: A 2-D boolean array the same size as dark and flat, with bad pixels marked as True.
    """
    if isinstance(master_dark, DetectorNoiseMaps):
        noise_map = master_dark
        master_dark = build_synthesized_dark(Dataset([master_flat]), noise_map)
        if dark_outputdir is not None:
            master_dark.save(filedir=dark_outputdir)

    #Extract data 
    dark_data = master_dark.data
    flat_data = master_flat.data

    #Calculate the hot and warm bad pixel maps
    hot_warm_pixels_bool = detect_hot_warm_pixels_from_dark(dark_data, dthresh)
    hot_warm_pixels = np.zeros_like(dark_data,dtype=np.uint8)
    hot_warm_pixels[hot_warm_pixels_bool] = 8 

    #Calculate the cold and dead pixels
    dead_pixels_bool = detect_dead_pixels_from_flat(flat_data, ffrac, fwidth)
    dead_pixels = np.zeros_like(dark_data,dtype=np.uint8) 
    dead_pixels[dead_pixels_bool] = 4

    #Combined the two maps 
    combined_badpixels = np.bitwise_or(hot_warm_pixels,dead_pixels)

    dark_hdr = master_dark.pri_hdr.copy()
    dark_ext_hdr = master_dark.ext_hdr.copy()

    dark_ext_hdr['DATALVL']  = 'CAL'
    dark_ext_hdr['HISTORY'] = "Bad Pixel Map created using {} and {}".format(master_dark.filename,master_flat.filename)

    input_dataset = Dataset([master_dark,master_flat])

    #Make the BadPixelMap object
    badpixelmap = BadPixelMap(combined_badpixels,pri_hdr=dark_hdr,ext_hdr=dark_ext_hdr,input_dataset=input_dataset)

    return badpixelmap
    
def detect_hot_warm_pixels_from_dark(dark, dthresh):
    """
    
    Detects warm/hot pixels from the dark frame, flagging anything above
    dthresh standard deviations above the mean dark level in the frame. For
    example, dthresh = 5 will flag any pixel > 5 standard deviations above the
    mean of dark.

    Ported from II&T pipeline

    Args:
        dark (array_like): 2-D array containing the master dark frame.
        dthresh (float): Threshold for flagging hot pixels, specified as the number of standard deviations above the mean dark level. Must be >= 0.

    Returns:
        fixedbp_dark (array_like): A 2-D boolean array the same size as the input dark frame, with hot pixels marked as True.
    """

    # Process dark frame
    fixedbp_dark = np.zeros(dark.shape).astype('bool')
    fixedbp_dark[dark > np.mean(dark) + dthresh*np.std(dark)] = True

    return fixedbp_dark

def detect_dead_pixels_from_flat(flat, ffrac, fwidth):
    """
    Compute a fixed bad pixel map from a flat field.

    Detects low- or non-functional pixels from the flat frame, flagging any
    pixels less than ffrac times the local mean flat level.  Flat uses
    local mean as flats may have low-spatial-frequency variations due to e.g.
    fringing or vignetting.  For example, ffrac = 0.8 and fwidth = 32 will
    flag any pixel which is < 80% of the mean value in a 32-pixel box
    centered on the pixel.  (Centration will use FFT rules, where odd-sized
    widths center on the pixel, and even-sized place the pixel to the right of
    center, e.g.:
     odd: [. x .]
     even: [. . x .]
    For boxes near the edge, only the subset of pixels within the box will be
    used for the calculation.

    Ported from II&T pipeline
    
    Args:
        flat (array_like): 2-D array containing the master flat field.
        ffrac (float): Fraction of the local mean value below which pixels are considered poorly-functioning.
        fwidth (int): Width of the box used to calculate the local mean, in pixels.

    Returns:
        fixedbp_flat (array_like): A 2-D boolean array the same size as the input flat frame, with dead pixels marked as True.

    """
    # Check inputs
    # check.twoD_array(flat, 'flat', TypeError)
    # check.real_nonnegative_scalar(ffrac, 'ffrac', TypeError)
    # check.positive_scalar_integer(fwidth, 'fwidth', TypeError)

    # Process flat frame
    fixedbp_flat = np.zeros(flat.shape).astype('bool')
    nrow, ncol = flat.shape
    for r in range(nrow):
        tmpr = np.arange(fwidth) - fwidth//2 + r
        rind = np.logical_and(tmpr >= 0, tmpr < nrow)
        for c in range(ncol):
            tmpc = np.arange(fwidth) - fwidth//2 + c
            cind = np.logical_and(tmpc >= 0, tmpc < ncol)

            # rind/cind removes indices that fall out of flat
            subflat = flat[tmpr[rind], :][:, tmpc[cind]]

            localm = np.mean(subflat)
            if flat[r, c] < ffrac*localm:
                fixedbp_flat[r, c] = True
                pass

            pass
        pass

    return fixedbp_flat



"""
Unit test suite for the calibrate_nonlin module.

Frames for unit tests are simulated SCI-size frames made with Gaussian and Poisson 
noises included. The assumed flux map is a realistic pupil image.
"""

import os
import pytest
import warnings
import numpy as np
import pandas as pd
from pathlib import Path

import test_check
import corgidrp
from corgidrp import check
from corgidrp.data import Image, Dataset, NonLinearityCalibration
from corgidrp.calibrate_nonlin import (calibrate_nonlin, CalNonlinException, nonlin_params_default)
from corgidrp.mocks import (make_fluxmap_image, nonlin_coefs)

def setup_module():
    """
    Sets module up and defines variables
    """
    global n_cal, n_mean
    global init_nonlins_arr
    global rms_test
    global exp_time_stack_arr0, time_stack_arr0, len_list0, gain_arr0
    global dataset_nl
    global norm_val, min_write, max_write

    ############################# prepare simulated frames #######################

    # path to nonlin table made from running calibrate_nonlin.py on TVAC frames
    # table used only to choose parameters to make analytic nonlin functions
    here = os.path.abspath(os.path.dirname(__file__))
    nonlin_table_path = Path(here,'test_data','nonlin_table_TVAC.txt')
    nonlin_flag = True # True adds nonlinearity to simulated frames

    # Load the arrays needed for calibrate_nonlin function from the .npz file
    loaded = np.load(Path(here,'test_data','nonlin_arrays_ut.npz'))
    # Access the arrays needed for calibrate_nonlin function
    exp_time_stack_arr0 = loaded['array1']
    time_stack_arr0 = loaded['array2']
    len_list0 = loaded['array3']
    gain_arr0 = loaded['array4'] # G = 1, 2, 10, 20
    # Reducing the number of frames used in unit tests (each has 5 substacks)
    n_cal = 1 
    idx_0 = 0
    idx_1 = 0
    # Turn into False and the test runs a similar amount of data as with real data in IIT (~40 Gb RAM)
    rms_test = 0.0035 # Value used when using an equivalent dataset as in IIT
    # Usual number of frames to deal with real rn values is ~200
    rn = 130 # read noise in e-
    if True:
        rms_test = 0.04 # Less strict comparison due to having significantly less frames
        # Usual number of frames to deal with real rn values is ~200
        rn = 130/np.sqrt(200/n_cal) # read noise in e-
        for iG in range(len(len_list0)):
            time_stack_arr0[idx_0:idx_0+n_cal*5] = time_stack_arr0[idx_1:idx_1+n_cal*5]
            for i_cal in range(n_cal):
                # There must be at least two frames with the same exposure time
                exp_time_stack_arr0[idx_0:idx_0+2] = exp_time_stack_arr0[idx_1]
                exp_time_stack_arr0[idx_0+2:idx_0+5] = exp_time_stack_arr0[idx_1]+ 0.05
                idx_0 += 5
            idx_1 += len_list0[iG]
        # Remove unused data
        exp_time_stack_arr0 = np.delete(exp_time_stack_arr0, idx_0 + np.arange(idx_1-idx_0))
        time_stack_arr0 = np.delete(time_stack_arr0, idx_0 + np.arange(idx_1-idx_0))
        # Update len_list0
        len_list0[:] = n_cal*5

    # Load the flux map
    fluxmap_init = np.load(Path(here,'test_data','FluxMap1024.npy'))
    fluxmap_init[fluxmap_init < 50] = 0 # cleanup flux map a bit
    fluxMap1 = 0.8*fluxmap_init # e/s/px, for G = 1
    fluxMap2 = 0.4*fluxmap_init # e/s/px, for G = 2
    fluxMap3 = 0.08*fluxmap_init # e/s/px, for G = 10
    fluxMap4 = 0.04*fluxmap_init # e/s/px, for G = 20

    # detector parameters
    kgain = 8.7 # e-/DN
    bias = 2000 # e-

    # cubic function nonlinearity for emgain of 1
    if nonlin_flag:
        coeffs_1, DNs, _ = nonlin_coefs(nonlin_table_path,1.0,3)
    else:
        coeffs_1 = [0.0, 0.0, 0.0, 1.0]
        _, DNs, _ = nonlin_coefs(nonlin_table_path,1.0,3)

    # commanded EM gain
    emgain = 1.0
    frame_list = []
    # make some uniform frames with emgain = 1 (must be unity)
    n_mean = 3
    for j in range(n_mean):
        image_sim = make_fluxmap_image(fluxMap1,bias,kgain,rn,emgain,5.0,coeffs_1,
            nonlin_flag=nonlin_flag)
        # Datetime cannot be duplicated
        image_sim.ext_hdr['DATETIME'] = time_stack_arr0[j]
        # Temporary keyword value. Mean frame is TBD
        image_sim.pri_hdr['OBSNAME'] = 'MNFRAME'
        frame_list.append(image_sim)

    init_nonlins = []
    index = 0
    for iG in range(len(gain_arr0)):
        g = gain_arr0[iG]
        exp_time_loop = exp_time_stack_arr0[index:index+len_list0[iG]]
        time_stack_test = time_stack_arr0[index:index+len_list0[iG]]
        index = index + len_list0[iG]
        if nonlin_flag:
            coeffs, _, vals = nonlin_coefs(nonlin_table_path,g,3)
        else:
            coeffs = [0.0, 0.0, 0.0, 1.0]
            vals = np.ones(len(DNs))
        init_nonlins.append(vals)
        for idx_t, t in enumerate(exp_time_loop):
            # Simulate full frame
            if iG == 0:
                image_sim = make_fluxmap_image(fluxMap1,bias,kgain,rn,g,t,coeffs,
                    nonlin_flag=nonlin_flag)
                image_sim.ext_hdr['DATETIME'] = time_stack_test[idx_t]
            elif iG == 1:
                image_sim = make_fluxmap_image(fluxMap2,bias,kgain,rn,g,t,coeffs,
                    nonlin_flag=nonlin_flag)
                image_sim.ext_hdr['DATETIME'] = time_stack_test[idx_t]
            elif iG == 2:
                image_sim = make_fluxmap_image(fluxMap3,bias,kgain,rn,g,t,coeffs,
                    nonlin_flag=nonlin_flag)
                image_sim.ext_hdr['DATETIME'] = time_stack_test[idx_t]
            else:
                image_sim = make_fluxmap_image(fluxMap4,bias,kgain,rn,g,t,coeffs,
                    nonlin_flag=nonlin_flag)
                image_sim.ext_hdr['DATETIME'] = time_stack_test[idx_t]
            image_sim.pri_hdr['OBSNAME'] = 'NONLIN'
            frame_list.append(image_sim)
    # Join all frames in a Dataset
    dataset_nl = Dataset(frame_list) 
    init_nonlins_arr = np.transpose(np.array(init_nonlins))

    # set input parameters for calibrate_nonlin
    local_path = os.path.dirname(os.path.realpath(__file__))
    exp_time_stack_arr = exp_time_stack_arr0
    time_stack_arr = time_stack_arr0
    len_list = len_list0
    norm_val = 3000
    min_write = 800
    max_write = 10000

def teardown_module():
    """
    Runs at the end. Deletes variables
    """
    global n_cal, n_mean
    global init_nonlins_arr
    global rms_test
    global exp_time_stack_arr0, time_stack_arr0, len_list0, gain_arr0
    global dataset_nl
    global norm_val, min_write, max_write

    del n_cal, n_mean
    del init_nonlins_arr
    del rms_test
    del exp_time_stack_arr0, time_stack_arr0, len_list0, gain_arr0
    del dataset_nl
    del norm_val, min_write, max_write

def test_expected_results_nom_sub():
    """Outputs are as expected for the provided frames with nominal arrays."""
    nonlin_out = calibrate_nonlin(dataset_nl, n_cal, n_mean, norm_val, min_write,
        max_write)

    # Test its DATALVL is CAL
    assert nonlin_out.ext_hdr['DATALVL'] == 'CAL'
        
    # Calculate rms of the differences between the assumed nonlinearity and 
    # the nonlinearity determined with calibrate_nonlin
    diffs0 = nonlin_out.data[1:,1] - init_nonlins_arr[:,0] # G = 1
    diffs1 = nonlin_out.data[1:,2] - init_nonlins_arr[:,1] # G = 2
    diffs2 = nonlin_out.data[1:,3] - init_nonlins_arr[:,2] # G = 10
    diffs3 = nonlin_out.data[1:,4] - init_nonlins_arr[:,3] # G = 20
    # Calculate rms
    rms1 = np.sqrt(np.mean(diffs0**2))
    rms2 = np.sqrt(np.mean(diffs1**2))
    rms3 = np.sqrt(np.mean(diffs2**2))
    rms4 = np.sqrt(np.mean(diffs3**2))

    # check that the four rms values are below some value (real data take
    # several frames and the value in IIT was 0.0035 for all of them)
    assert np.less(rms1,rms_test)
    assert np.less(rms2,rms_test)
    assert np.less(rms3,rms_test)
    assert np.less(rms4,rms_test)
    # check that the first element in the first column is equal to min_write
    assert np.equal(nonlin_out.data[1,0], min_write)
    # check that the last element in the first column is equal to max_write
    assert np.equal(nonlin_out.data[-1,0], max_write)
    # check that the unity value is in the correct row
    norm_ind = np.where(nonlin_out.data[1:, 1] == 1)[0]
    assert np.equal(nonlin_out.data[norm_ind+1,1], 1)
    assert np.equal(nonlin_out.data[norm_ind+1,-1], 1)
    # check that norm_val is correct
    assert np.equal(nonlin_out.data[norm_ind+1,0], norm_val)

    # Test filename follows convention (as of R3.0.2)
    datadir = os.path.join(os.path.dirname(__file__), "simdata")
    if not os.path.exists(datadir):
        os.mkdir(datadir)
    nonlin_out.save(filedir=datadir)
    nln_cal_filename = dataset_nl[-1].filename.replace("_L2b", "_NLN_CAL")
    nln_cal_filepath = os.path.join(datadir, nln_cal_filename)
    if os.path.exists(nln_cal_filepath) is False:
        raise IOError(f'NonLinearity calibration file {nln_cal_filepath} does not exist.')
    # Load the calibration file to check it has the same data contents
    nln_cal_file_load = NonLinearityCalibration(nln_cal_filepath)
    # The first element is a NaN
    assert np.isnan(nln_cal_file_load.data[0,0])
    # Compare all other elements but for [0,0], which is a NaN
    for row in range(nln_cal_file_load.data.shape[0]):
        for col in range(nln_cal_file_load.data.shape[1]):
            if row == 0 and col == 0:
                continue
            assert np.all(nln_cal_file_load.data[row,col] == nonlin_out.data[row,col])

    # Remove test NLN cal file
    if os.path.exists(nonlin_out.filepath):
        os.remove(nonlin_out.filepath)

def test_expected_results_time_sub():
    """Outputs are as expected for the provided frames with datetime values for
    one EM gain group taken 1 day later. Set (gain_arr0[3]) taken 1 day later
    avoiding a duplication of the whole dataset."""
    time_stack_arr1 = np.copy(time_stack_arr0)
    # Convert date-time strings to datetime objects. First ones are used to create a mean frame
    index = len(dataset_nl) - n_mean - len_list0[3]
    ctime_datetime1 = pd.to_datetime(time_stack_arr1[index:], errors='coerce')
    # Add one day to each datetime object
    ctime_datetime_plus_one_day1 = ctime_datetime1 + pd.Timedelta(days=1)
    # Convert back to strings
    ctime_strings_plus_one_day1 = ctime_datetime_plus_one_day1.strftime('%Y-%m-%dT%H:%M:%S').tolist()
    time_stack_arr1[index:index+len_list0[3]] = ctime_strings_plus_one_day1
    index = 0
    # First ones are used to create a mean frame
    idx_frame = n_mean
    for iG in range(len(gain_arr0)):
        exp_time_loop = exp_time_stack_arr0[index:index+len_list0[iG]]
        time_stack_test = time_stack_arr1[index:index+len_list0[iG]]
        index = index + len_list0[iG]
        for idx_t, t in enumerate(exp_time_loop):
            dataset_nl[idx_frame].ext_hdr['DATETIME'] = time_stack_test[idx_t]
            idx_frame += 1

    nonlin_out = calibrate_nonlin(dataset_nl, n_cal, n_mean, norm_val, min_write, max_write)
     
    # Calculate rms of the differences between the assumed nonlinearity and 
    # the nonlinearity determined with calibrate_nonlin
    diffs0 = nonlin_out.data[1:,1] - init_nonlins_arr[:,0] # G = 1
    diffs1 = nonlin_out.data[1:,2] - init_nonlins_arr[:,1] # G = 2
    diffs2 = nonlin_out.data[1:,3] - init_nonlins_arr[:,2] # G = 10
    diffs3 = nonlin_out.data[1:,4] - init_nonlins_arr[:,3] # G = 20
    # Calculte rms and peak-to-peak differences
    rms1 = np.sqrt(np.mean(diffs0**2))
    rms2 = np.sqrt(np.mean(diffs1**2))
    rms3 = np.sqrt(np.mean(diffs2**2))
    rms4 = np.sqrt(np.mean(diffs3**2))
    
    # check that the four rms values are below some value (real data take
    # several frames and the value in IIT was 0.0035 for all of them)
    assert np.less(rms1,rms_test)
    assert np.less(rms2,rms_test)
    assert np.less(rms3,rms_test)
    assert np.less(rms4,rms_test)
  

def teardown_module():
    """
    Run at end of tests. Deletes variables

    """
    global n_cal, n_mean
    global init_nonlins_arr
    global rms_test
    global exp_time_stack_arr0, time_stack_arr0, len_list0, gain_arr0
    global dataset_nl

    del n_cal, n_mean
    del init_nonlins_arr
    del rms_test
    del exp_time_stack_arr0, time_stack_arr0, len_list0, gain_arr0
    del dataset_nl


def test_norm_val():
    """norm_val must be divisible by 20."""
    norm_not_div_20 = 2010
    with pytest.raises(CalNonlinException):
        calibrate_nonlin(dataset_nl, n_cal, n_mean, norm_not_div_20, min_write, max_write)
 
def test_rps():
    """these two below must be real positive scalars."""
    check_list = test_check.rpslist
    # min_write
    for rerr in check_list:
        with pytest.raises(TypeError):
            calibrate_nonlin(dataset_nl, n_cal, n_mean, norm_val, rerr, max_write)
    # max_write
    for rerr in check_list:
        with pytest.raises(TypeError):
            calibrate_nonlin(dataset_nl, n_cal, n_mean, norm_val, min_write, rerr)
   
def test_max_gt_min():
    """max_write must be greater than min_write."""
    werr = min_write # set the max_write value to the min_write value
    with pytest.raises(CalNonlinException):
        calibrate_nonlin(dataset_nl, n_cal, n_mean, norm_val, min_write, werr)
    
def test_psi():
    """norm_val must be a positive scalar integer."""
    check_list = test_check.psilist
    # norm_val
    for mmerr in check_list:
        with pytest.raises(TypeError):
            calibrate_nonlin(dataset_nl, n_cal, n_mean, mmerr, min_write, max_write)

def test_nonlin_params():
    '''Check that the check function for nonlin_params works as expected.'''
    # test bad input for nonlin_params
    nonlin_params_bad = nonlin_params_default.copy()
    nonlin_params_bad['colroi2'] = 'foo'
    with pytest.raises(TypeError):
        calibrate_nonlin(dataset_nl, n_cal, n_mean, norm_val, min_write, max_write,
                        nonlin_params=nonlin_params_bad)
 
if __name__ == '__main__':
    print('Running test_nonlin_params')
    test_nonlin_params()
    print('Running test_expected_results_nom_sub')
    test_expected_results_nom_sub()
    print('Running test_expected_results_time_sub')
    test_expected_results_time_sub()
    print('Running test_norm_val')
    test_norm_val()
    print('Running test_rps')
    test_rps()
    print('Running test_max_gt_min')
    test_max_gt_min()
    print('Running test_psi')
    test_psi()
    print('Non-Linearity Calibration tests passed')



import copy
import numpy as np

import corgidrp.data as data

def extract_frame_id(filename):
    """
    Extract frame ID from an L1 filename. Structure is assumed to be ending
    like '..._frame_id.fits' where frame_id is a series of digits

    Args:
      filename: L1 filename

    Returns:
      Frame id as a string of length 10
    """
    idx_0 = len(filename) - filename[::-1].find('_')
    idx_1 = len(filename) - filename[::-1].find('.') - 1

    return int(filename[idx_0:idx_1])

def sort_pupilimg_frames(
    dataset_in,
    cal_type=''):
    """ 
    Sorting algorithm that given a dataset will output a dataset with the
    frames used to generate a mean frame and the frames used to calibrate
    the calibration type: k-gain. non-linearity.

    The output dataset has an added keyword value in its extended header:
    OBSNAME with values 'MNFRAME' (for mean frame), 'KGAIN' (for K-gain),
    and 'NONLIN' (for non-linearity).

    Args:
      dataset_in (corgidrp.Dataset): dataset with all the frames to be sorted.
        By default, it is expected to contain all the frames from the PUPILIMG
        visit files associated with a calibration campaign
      cal_type (string): the calibration type. Case insensitive.
        Accepted values are:
        'k-gain' or 'kgain' for K-gain calibration
        'non-lin(earity)', 'nonlin(earity)' for non-linearity calibration, where
        the letters within parenthesis are optional.

    Returns:

        Dataset with the frames used to generate a mean frame and the frames
        used to calibrate the calibration type: k-gain or non-linearity.
    """
    # Copy dataset
    dataset_cp = dataset_in.copy()
    # Split by CMDGAIN
    split_cmdgain = dataset_cp.split_dataset(exthdr_keywords=['EMGAIN_C'])
    # Mean frame: split by EXPTIME
    idx_unity = np.where(np.array(split_cmdgain[1])==1)[0][0]
    split_exptime = split_cmdgain[0][idx_unity].split_dataset(exthdr_keywords=['EXPTIME'])
    # Get the set with the most frames
    n_frames_list = np.zeros(len(split_exptime[0]))
    for i_sub, subset in enumerate(split_exptime[0]):
        n_frames_list[i_sub] = len(split_exptime[0][i_sub])
    # Choice: choose the subset with the maximum number of frames
    idx_mean_frame = np.argmax(n_frames_list)
    frame_id_list = []
    for frame in split_exptime[0][idx_mean_frame]:
        frame_id_list += [extract_frame_id(frame.filename)]
    # Choose the frames with consecutive ID numbers (same row in AUX file)
    frame_id_sort = np.array(frame_id_list)
    frame_id_sort.sort()
    count_cons = [1]
    idx_cons = 0
    for idx in range(len(frame_id_sort)-1):
        if frame_id_sort[idx+1] - frame_id_sort[idx] == 1:
            count_cons[idx_cons] += 1
        else:
            idx_cons += 1
            count_cons += [1]
    # Choose the largest subset
    idx_mean_frame_cons = np.argmax(count_cons)
    idx_mean_frame_last = np.sum(count_cons[0:idx_mean_frame_cons+1]).astype(int)
    idx_mean_frame_first = idx_mean_frame_last - count_cons[idx_mean_frame_cons]
    frame_id_mean_frame = frame_id_sort[idx_mean_frame_first:idx_mean_frame_last]
    mean_frame_list = []

    n_mean_frame = 0
    for frame in split_exptime[0][idx_mean_frame]:
        if int(extract_frame_id(frame.filename)) in frame_id_mean_frame:
            exptime_mean_frame = frame.ext_hdr['EXPTIME']
            # Update keyword OBSNAME
            frame.pri_hdr['OBSNAME'] = 'MNFRAME'
            mean_frame_list += [frame]
            n_mean_frame += 1
            
    sorting_summary = (f'Mean frame has {n_mean_frame} unity frames with' + 
        f' exposure time {exptime_mean_frame} seconds. ')

    # K-gain and non-linearity
    cal_frame_list = []
    if cal_type.lower() == 'k-gain' or cal_type.lower() == 'kgain':
        print('Considering K-gain:')
    elif cal_type.lower()[0:7] == 'non-lin' or cal_type.lower()[0:6] == 'nonlin':
        print('Considering Non-linearity:')
    else:
        raise Exception('Unrecognized calibration type (expected k-gain, non-lin)')

    # Remove main frame frames from unity gain frames
    split_exptime[0].remove(split_exptime[0][idx_mean_frame])
    split_exptime[1].remove(split_exptime[1][idx_mean_frame])
    # Frames must be taken consecutively
    frame_id_list = []
    exptime_list = []
    unity_gain_filepath_list = []
    for subset in split_exptime[0]:
        for frame in subset:
            frame_id_list += [extract_frame_id(frame.filename)]
            exptime_list += [frame.ext_hdr['EXPTIME']]
            unity_gain_filepath_list += [frame.filepath]
    idx_id_sort = np.argsort(frame_id_list)
    exptime_arr = np.array(exptime_list)[idx_id_sort]
    # Count repeated, consecutive elements
    count_cons = [1]
    exptime_cons = [exptime_arr[0]]
    idx_cons = 0
    for exptime in exptime_arr:
        if exptime == exptime_cons[-1]:
            count_cons[idx_cons] += 1
        else:
            idx_cons += 1
            count_cons += [1]
            exptime_cons += [exptime]
    # First index always has a repetition in the previous loop (id=id)
    count_cons[0] -= 1

    idx_cons2 = [0]
    exptime_cons2 = [exptime_cons[idx_cons2[0]]]
    kgain_subset = []
    # Iterate over unique counts that are consecutive
    for idx_count in range(len(count_cons) - 1):
        # Indices to cover two consecutive sets
        idx_id_first = np.sum(count_cons[0:idx_count]).astype(int)
        idx_id_last  = np.sum(count_cons[0:idx_count+2]).astype(int)
        diff_id = np.diff(np.array(frame_id_list)[idx_id_sort[idx_id_first:idx_id_last]])
        diff_exp = np.diff(exptime_cons)
        # Both subsets must have all Ids consecutive because they are in
        # time order
        if (count_cons[idx_count+1] == count_cons[idx_count] and
            np.all(diff_id == 1) and diff_exp[idx_count] > 0):
            exptime_cons2 += [exptime_cons[idx_count+1]]
            idx_cons2 += [idx_count+1]
        # Last exposure time must be repeated and only once
        elif (diff_exp[idx_count] < 0  and
            exptime_cons[idx_count+1] in exptime_cons[0:idx_count+1] and
            len(exptime_cons2) == len(set(exptime_cons2))):
            kgain_subset += [idx_cons2[0], idx_count+1]
            idx_cons2 = [idx_count+1]
            exptime_cons2 = [exptime_cons]
        else:
        # It is not a subset for kgain
           continue
    # Choose the largest subset
    kgain_subset = np.array(kgain_subset)
    idx_kgain = np.argmax(kgain_subset[1::2] - kgain_subset[0::2])
    # Extract first/last index in the subset of consecutive frames
    idx_kgain_0 = kgain_subset[2*idx_kgain]
    idx_kgain_1 = kgain_subset[2*idx_kgain + 1]
    # Count frames before and subset length
    idx_kgain_first = np.sum(count_cons[0:idx_kgain_0]).astype(int)
    idx_kgain_last = (idx_kgain_first +
        np.sum(count_cons[idx_kgain_0:idx_kgain_1+1]).astype(int))

    # Sort unity gain filenames
    unity_gain_filepath_arr = np.array(unity_gain_filepath_list)[idx_id_sort]
    cal_list = unity_gain_filepath_arr[idx_kgain_first:idx_kgain_last]
    # Update OBSNAME and take profit to check files are in the list
    n_kgain = 0
    cal_frame_list = []
    for frame in dataset_cp:
        if frame.filepath in cal_list:
            vistype = frame.pri_hdr['VISTYPE']
            frame.pri_hdr['OBSNAME'] = 'KGAIN'
            cal_frame_list += [frame]
            n_kgain += 1

    sorting_summary += (f'K-gain has {n_kgain} unity frames with exposure ' +
        f'times {exptime_cons[idx_kgain_0:idx_kgain_1+1]} seconds with ' +
        f'{count_cons[idx_kgain_0]} frames each. ')

    # Non-unity gain frames for Non-linearity
    if cal_type.lower()[0:7] == 'non-lin' or cal_type.lower()[0:6] == 'nonlin':
        # Non-unity gain frames
        split_cmdgain[0].remove(split_cmdgain[0][idx_unity])
        split_cmdgain[1].remove(split_cmdgain[1][idx_unity])
        n_nonlin = 0
        nonlin_emgain = []
        for idx_gain_set, gain_set in enumerate(split_cmdgain[0]):
            # Frames must be taken consecutively
            frame_id_list = []
            exptime_list = []
            gain_filepath_list = []
            for frame in gain_set:
                frame_id_list += [extract_frame_id(frame.filename)]
                exptime_list += [frame.ext_hdr['EXPTIME']]
                gain_filepath_list += [frame.filepath]
            # One can set a stronger condition, though in the end the max set
            if len(frame_id_list) < 3:
                continue
            idx_id_sort = np.argsort(frame_id_list)
            exptime_arr = np.array(exptime_list)[idx_id_sort]
            # We need an increasing series of exposure times with the last one
            # the only repeated value in the series
            idx_subsets = np.where(np.diff(exptime_arr) < 0)[0]
            if len(idx_subsets) == 0:
                continue
            # length of candidate subsets
            nonlin_len = []
            for idx, idx_subset in enumerate(idx_subsets):
                # Add 0th element plus the one lost in diff
                if idx == 0:
                    exptime_tmp = exptime_arr[0:idx_subset+2]
                else:
                    exptime_tmp = exptime_arr[idx_subsets[idx-1]+1:idx_subset+2]
                # Check conditions
                if (exptime_tmp[-1] in exptime_tmp[:-1] and
                    len(exptime_tmp) - 1 == len(set(exptime_tmp))):
                    nonlin_len += [len(exptime_tmp)]
                else:
                    nonlin_len += [-1]
            # COntinue of there are no good candidates
            if np.max(nonlin_len) <= 0:
                continue
            # Find maximum set among good candidates
            if np.argmax(nonlin_len) == 0:
                idx_nonlin_first = 0
                idx_nonlin_last = idx_subsets[0] + 1
            else:
                idx_nonlin_first = idx_subsets[np.argmax(nonlin_len)-1] + 1
                idx_nonlin_last = idx_subsets[np.argmax(nonlin_len)] + 1
            # Sort unity gain filenames
            gain_filepath_arr = np.array(gain_filepath_list)[idx_id_sort]
            cal_list = gain_filepath_arr[idx_nonlin_first:idx_nonlin_last+1]
            # Update OBSNAME and take profit to check files are in the list
            for frame in dataset_cp:
                if frame.filepath in cal_list:
                    vistype = frame.pri_hdr['VISTYPE']
                    frame.pri_hdr['OBSNAME'] = 'NONLIN'
                    cal_frame_list += [frame]
                    n_nonlin += 1
            nonlin_emgain += [split_cmdgain[1][idx_gain_set]]

        sorting_summary += (f'Non-linearity has {n_nonlin} frames with gains ' +
            f'{nonlin_emgain}')
        
    # TODO: Add a HISTORY entry
    history = (f'Dataset to calibrate {cal_type.upper()}. A sorting algorithm ' +
        'based on the constraints that NFRAMES, EXPTIME and CMDGAIN have when collecting ' +
        'calibration data for K-gain, Non-linearity and EM-gain vs DAC '
        f"was applied to an input dataset from {vistype} visit files." +
        f'The result is that {sorting_summary}.')
    print(history)

    dataset_sorted = data.Dataset(mean_frame_list + cal_frame_list)
    dataset_sorted.update_after_processing_step(history)
    # Return Datafrane with mean frame and cal type
    return dataset_sorted
